import os
import platform
import sys
import warnings
from collections import deque
from contextlib import contextmanager
from copy import copy, deepcopy
from datetime import datetime, timezone
from enum import Enum
from functools import wraps
from itertools import chain
from typing import TYPE_CHECKING, cast

import sentry_sdk
from sentry_sdk._types import AnnotatedValue
from sentry_sdk.attachments import Attachment
from sentry_sdk.consts import (
    DEFAULT_MAX_BREADCRUMBS,
    FALSE_VALUES,
    INSTRUMENTER,
    SPANDATA,
)
from sentry_sdk.feature_flags import DEFAULT_FLAG_CAPACITY, FlagBuffer
from sentry_sdk.profiler.continuous_profiler import (
    get_profiler_id,
    try_autostart_continuous_profiler,
    try_profile_lifecycle_trace_start,
)
from sentry_sdk.profiler.transaction_profiler import Profile
from sentry_sdk.session import Session
from sentry_sdk.traces import (
    _DEFAULT_PARENT_SPAN,
    NoOpStreamedSpan,
    StreamedSpan,
)
from sentry_sdk.tracing import (
    BAGGAGE_HEADER_NAME,
    SENTRY_TRACE_HEADER_NAME,
    NoOpSpan,
    Span,
    Transaction,
)
from sentry_sdk.tracing_utils import (
    Baggage,
    PropagationContext,
    _make_sampling_decision,
    has_span_streaming_enabled,
    has_tracing_enabled,
    is_ignored_span,
)
from sentry_sdk.utils import (
    ContextVar,
    capture_internal_exception,
    capture_internal_exceptions,
    datetime_from_isoformat,
    disable_capture_event,
    event_from_exception,
    exc_info_from_error,
    format_attribute,
    has_logs_enabled,
    has_metrics_enabled,
    logger,
)

if TYPE_CHECKING:
    from collections.abc import Mapping
    from typing import (
        Any,
        Callable,
        Deque,
        Dict,
        Generator,
        Iterator,
        List,
        Optional,
        ParamSpec,
        Tuple,
        TypeVar,
        Union,
    )

    from typing_extensions import Unpack

    import sentry_sdk
    from sentry_sdk._types import (
        Attributes,
        AttributeValue,
        Breadcrumb,
        BreadcrumbHint,
        ErrorProcessor,
        Event,
        EventProcessor,
        ExcInfo,
        Hint,
        Log,
        LogLevelStr,
        Metric,
        SamplingContext,
        Type,
    )
    from sentry_sdk.tracing import TransactionKwargs

    P = ParamSpec("P")
    R = TypeVar("R")

    F = TypeVar("F", bound=Callable[..., Any])
    T = TypeVar("T")


# Holds data that will be added to **all** events sent by this process.
# In case this is a http server (think web framework) with multiple users
# the data will be added to events of all users.
# Typically this is used for process wide data such as the release.
_global_scope: "Optional[Scope]" = None

# Holds data for the active request.
# This is used to isolate data for different requests or users.
# The isolation scope is usually created by integrations, but may also
# be created manually
_isolation_scope = ContextVar("isolation_scope", default=None)

# Holds data for the active span.
# This can be used to manually add additional data to a span.
_current_scope = ContextVar("current_scope", default=None)

global_event_processors: "List[EventProcessor]" = []

# A function returning a (trace_id, span_id) tuple
# from an external tracing source (such as otel)
_external_propagation_context_fn: "Optional[Callable[[], Optional[Tuple[str, str]]]]" = None


class ScopeType(Enum):
    CURRENT = "current"
    ISOLATION = "isolation"
    GLOBAL = "global"
    MERGED = "merged"


class _ScopeManager:
    def __init__(self, hub: "Optional[Any]" = None) -> None:
        self._old_scopes: "List[Scope]" = []

    def __enter__(self) -> "Scope":
        isolation_scope = Scope.get_isolation_scope()

        self._old_scopes.append(isolation_scope)

        forked_scope = isolation_scope.fork()
        _isolation_scope.set(forked_scope)

        return forked_scope

    def __exit__(self, exc_type: "Any", exc_value: "Any", tb: "Any") -> None:
        old_scope = self._old_scopes.pop()
        _isolation_scope.set(old_scope)


def add_global_event_processor(processor: "EventProcessor") -> None:
    global_event_processors.append(processor)


def register_external_propagation_context(
    fn: "Callable[[], Optional[Tuple[str, str]]]",
) -> None:
    global _external_propagation_context_fn
    _external_propagation_context_fn = fn


def remove_external_propagation_context() -> None:
    global _external_propagation_context_fn
    _external_propagation_context_fn = None


def get_external_propagation_context() -> "Optional[Tuple[str, str]]":
    return (
        _external_propagation_context_fn() if _external_propagation_context_fn else None
    )


def has_external_propagation_context() -> bool:
    return _external_propagation_context_fn is not None


def _attr_setter(fn: "Any") -> "Any":
    return property(fset=fn, doc=fn.__doc__)


def _disable_capture(fn: "F") -> "F":
    @wraps(fn)
    def wrapper(self: "Any", *args: "Dict[str, Any]", **kwargs: "Any") -> "Any":
        if not self._should_capture:
            return
        try:
            self._should_capture = False
            return fn(self, *args, **kwargs)
        finally:
            self._should_capture = True

    return wrapper  # type: ignore


class Scope:
    """The scope holds extra information that should be sent with all
    events that belong to it.
    """

    # NOTE: Even though it should not happen, the scope needs to not crash when
    # accessed by multiple threads. It's fine if it's full of races, but those
    # races should never make the user application crash.
    #
    # The same needs to hold for any accesses of the scope the SDK makes.

    __slots__ = (
        "_level",
        "_name",
        "_fingerprint",
        # note that for legacy reasons, _transaction is the transaction *name*,
        # not a Transaction object (the object is stored in _span)
        "_transaction",
        "_transaction_info",
        "_user",
        "_tags",
        "_contexts",
        "_extras",
        "_breadcrumbs",
        "_n_breadcrumbs_truncated",
        "_gen_ai_original_message_count",
        "_gen_ai_conversation_id",
        "_event_processors",
        "_error_processors",
        "_should_capture",
        "_span",
        "_session",
        "_attachments",
        "_force_auto_session_tracking",
        "_profile",
        "_propagation_context",
        "client",
        "_type",
        "_last_event_id",
        "_flags",
        "_attributes",
    )

    def __init__(
        self,
        ty: "Optional[ScopeType]" = None,
        client: "Optional[sentry_sdk.Client]" = None,
    ) -> None:
        self._type = ty

        self._event_processors: "List[EventProcessor]" = []
        self._error_processors: "List[ErrorProcessor]" = []

        self._name: "Optional[str]" = None
        self._propagation_context: "Optional[PropagationContext]" = None
        self._n_breadcrumbs_truncated: int = 0
        self._gen_ai_original_message_count: "Dict[str, int]" = {}

        self.client: "sentry_sdk.client.BaseClient" = NonRecordingClient()

        if client is not None:
            self.set_client(client)

        self.clear()

        incoming_trace_information = self._load_trace_data_from_env()
        self.generate_propagation_context(incoming_data=incoming_trace_information)

    def __copy__(self) -> "Scope":
        """
        Returns a copy of this scope.
        This also creates a copy of all referenced data structures.
        """
        rv: "Scope" = object.__new__(self.__class__)

        rv._type = self._type
        rv.client = self.client
        rv._level = self._level
        rv._name = self._name
        rv._fingerprint = self._fingerprint
        rv._transaction = self._transaction
        rv._transaction_info = self._transaction_info.copy()
        rv._user = self._user

        rv._tags = self._tags.copy()
        rv._contexts = self._contexts.copy()
        rv._extras = self._extras.copy()

        rv._breadcrumbs = copy(self._breadcrumbs)
        rv._n_breadcrumbs_truncated = self._n_breadcrumbs_truncated
        rv._gen_ai_original_message_count = self._gen_ai_original_message_count.copy()
        rv._event_processors = self._event_processors.copy()
        rv._error_processors = self._error_processors.copy()
        rv._propagation_context = self._propagation_context

        rv._should_capture = self._should_capture
        rv._span = self._span
        rv._session = self._session
        rv._force_auto_session_tracking = self._force_auto_session_tracking
        rv._attachments = self._attachments.copy()

        rv._profile = self._profile

        rv._last_event_id = self._last_event_id

        rv._flags = deepcopy(self._flags)

        rv._attributes = self._attributes.copy()

        rv._gen_ai_conversation_id = self._gen_ai_conversation_id

        return rv

    @classmethod
    def get_current_scope(cls) -> "Scope":
        """
        .. versionadded:: 2.0.0

        Returns the current scope.
        """
        current_scope = _current_scope.get()
        if current_scope is None:
            current_scope = Scope(ty=ScopeType.CURRENT)
            _current_scope.set(current_scope)

        return current_scope

    @classmethod
    def set_current_scope(cls, new_current_scope: "Scope") -> None:
        """
        .. versionadded:: 2.0.0

        Sets the given scope as the new current scope overwriting the existing current scope.
        :param new_current_scope: The scope to set as the new current scope.
        """
        _current_scope.set(new_current_scope)

    @classmethod
    def get_isolation_scope(cls) -> "Scope":
        """
        .. versionadded:: 2.0.0

        Returns the isolation scope.
        """
        isolation_scope = _isolation_scope.get()
        if isolation_scope is None:
            isolation_scope = Scope(ty=ScopeType.ISOLATION)
            _isolation_scope.set(isolation_scope)

        return isolation_scope

    @classmethod
    def set_isolation_scope(cls, new_isolation_scope: "Scope") -> None:
        """
        .. versionadded:: 2.0.0

        Sets the given scope as the new isolation scope overwriting the existing isolation scope.
        :param new_isolation_scope: The scope to set as the new isolation scope.
        """
        _isolation_scope.set(new_isolation_scope)

    @classmethod
    def get_global_scope(cls) -> "Scope":
        """
        .. versionadded:: 2.0.0

        Returns the global scope.
        """
        global _global_scope
        if _global_scope is None:
            _global_scope = Scope(ty=ScopeType.GLOBAL)

        return _global_scope

    def set_global_attributes(self) -> None:
        from sentry_sdk.client import SDK_INFO

        self.set_attribute(SPANDATA.SENTRY_SDK_NAME, SDK_INFO["name"])
        self.set_attribute(SPANDATA.SENTRY_SDK_VERSION, SDK_INFO["version"])

        self.set_attribute(
            "process.runtime.name",
            platform.python_implementation(),
        )
        self.set_attribute(
            "process.runtime.version",
            f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
        )

        options = sentry_sdk.get_client().options

        server_name = options.get("server_name")
        if server_name:
            self.set_attribute(SPANDATA.SERVER_ADDRESS, server_name)

        environment = options.get("environment")
        if environment:
            self.set_attribute(SPANDATA.SENTRY_ENVIRONMENT, environment)

        release = options.get("release")
        if release:
            self.set_attribute(SPANDATA.SENTRY_RELEASE, release)

    @classmethod
    def last_event_id(cls) -> "Optional[str]":
        """
        .. versionadded:: 2.2.0

        Returns event ID of the event most recently captured by the isolation scope, or None if no event
        has been captured. We do not consider events that are dropped, e.g. by a before_send hook.
        Transactions also are not considered events in this context.

        The event corresponding to the returned event ID is NOT guaranteed to actually be sent to Sentry;
        whether the event is sent depends on the transport. The event could be sent later or not at all.
        Even a sent event could fail to arrive in Sentry due to network issues, exhausted quotas, or
        various other reasons.
        """
        return cls.get_isolation_scope()._last_event_id

    def _merge_scopes(
        self,
        additional_scope: "Optional[Scope]" = None,
        additional_scope_kwargs: "Optional[Dict[str, Any]]" = None,
    ) -> "Scope":
        """
        Merges global, isolation and current scope into a new scope and
        adds the given additional scope or additional scope kwargs to it.
        """
        if additional_scope and additional_scope_kwargs:
            raise TypeError("cannot provide scope and kwargs")

        final_scope = copy(_global_scope) if _global_scope is not None else Scope()
        final_scope._type = ScopeType.MERGED

        isolation_scope = _isolation_scope.get()
        if isolation_scope is not None:
            final_scope.update_from_scope(isolation_scope)

        current_scope = _current_scope.get()
        if current_scope is not None:
            final_scope.update_from_scope(current_scope)

        if self != current_scope and self != isolation_scope:
            final_scope.update_from_scope(self)

        if additional_scope is not None:
            if callable(additional_scope):
                additional_scope(final_scope)
            else:
                final_scope.update_from_scope(additional_scope)

        elif additional_scope_kwargs:
            final_scope.update_from_kwargs(**additional_scope_kwargs)

        return final_scope

    @classmethod
    def get_client(cls) -> "sentry_sdk.client.BaseClient":
        """
        .. versionadded:: 2.0.0

        Returns the currently used :py:class:`sentry_sdk.Client`.
        This checks the current scope, the isolation scope and the global scope for a client.
        If no client is available a :py:class:`sentry_sdk.client.NonRecordingClient` is returned.
        """
        current_scope = _current_scope.get()
        try:
            client = current_scope.client
        except AttributeError:
            client = None

        if client is not None and client.is_active():
            return client

        isolation_scope = _isolation_scope.get()
        try:
            client = isolation_scope.client
        except AttributeError:
            client = None

        if client is not None and client.is_active():
            return client

        try:
            client = _global_scope.client  # type: ignore
        except AttributeError:
            client = None

        if client is not None and client.is_active():
            return client

        return NonRecordingClient()

    def set_client(
        self, client: "Optional[sentry_sdk.client.BaseClient]" = None
    ) -> None:
        """
        .. versionadded:: 2.0.0

        Sets the client for this scope.

        :param client: The client to use in this scope.
            If `None` the client of the scope will be replaced by a :py:class:`sentry_sdk.NonRecordingClient`.

        """
        if client is not None:
            self.client = client
            # We need a client to set the initial global attributes on the global
            # scope since they mostly come from client options, so populate them
            # as soon as a client is set
            sentry_sdk.get_global_scope().set_global_attributes()
        else:
            self.client = NonRecordingClient()

    def fork(self) -> "Scope":
        """
        .. versionadded:: 2.0.0

        Returns a fork of this scope.
        """
        forked_scope = copy(self)
        return forked_scope

    def _load_trace_data_from_env(self) -> "Optional[Dict[str, str]]":
        """
        Load Sentry trace id and baggage from environment variables.
        Can be disabled by setting SENTRY_USE_ENVIRONMENT to "false".
        """
        incoming_trace_information: "Optional[Dict[str, str]]" = None

        sentry_use_environment = (
            os.environ.get("SENTRY_USE_ENVIRONMENT") or ""
        ).lower()
        use_environment = sentry_use_environment not in FALSE_VALUES
        if use_environment:
            incoming_trace_information = {}

            if os.environ.get("SENTRY_TRACE"):
                incoming_trace_information[SENTRY_TRACE_HEADER_NAME] = (
                    os.environ.get("SENTRY_TRACE") or ""
                )

            if os.environ.get("SENTRY_BAGGAGE"):
                incoming_trace_information[BAGGAGE_HEADER_NAME] = (
                    os.environ.get("SENTRY_BAGGAGE") or ""
                )

        return incoming_trace_information or None

    def set_new_propagation_context(self) -> None:
        """
        Creates a new propagation context and sets it as `_propagation_context`. Overwriting existing one.
        """
        self._propagation_context = PropagationContext()

    def generate_propagation_context(
        self, incoming_data: "Optional[Dict[str, str]]" = None
    ) -> None:
        """
        Makes sure the propagation context is set on the scope.
        If there is `incoming_data` overwrite existing propagation context.
        If there is no `incoming_data` create new propagation context, but do NOT overwrite if already existing.
        """
        if incoming_data is not None:
            self._propagation_context = PropagationContext.from_incoming_data(
                incoming_data
            )

        # TODO-neel this below is a BIG code smell but requires a bunch of other refactoring
        if self._type != ScopeType.CURRENT:
            if self._propagation_context is None:
                self.set_new_propagation_context()

    def get_dynamic_sampling_context(self) -> "Optional[Dict[str, str]]":
        """
        Returns the Dynamic Sampling Context from the Propagation Context.
        If not existing, creates a new one.

        Deprecated: Logic moved to PropagationContext, don't use directly.
        """
        if self._propagation_context is None:
            return None

        return self._propagation_context.dynamic_sampling_context

    def get_traceparent(self, *args: "Any", **kwargs: "Any") -> "Optional[str]":
        """
        Returns the Sentry "sentry-trace" header (aka the traceparent) from the
        currently active span or the scopes Propagation Context.
        """
        client = self.get_client()

        if not has_tracing_enabled(client.options):
            return self.get_active_propagation_context().to_traceparent()

        span_streaming = has_span_streaming_enabled(client.options)
        # If we have an active span, return traceparent from there
        if span_streaming and self.streamed_span is not None:
            return self.streamed_span._to_traceparent()
        elif not span_streaming and self.span is not None:
            return self.span._to_traceparent()

        # else return traceparent from the propagation context
        return self.get_active_propagation_context().to_traceparent()

    def get_baggage(self, *args: "Any", **kwargs: "Any") -> "Optional[Baggage]":
        """
        Returns the Sentry "baggage" header containing trace information from the
        currently active span or the scopes Propagation Context.
        """
        client = self.get_client()

        if not has_tracing_enabled(client.options):
            return self.get_active_propagation_context().get_baggage()

        span_streaming = has_span_streaming_enabled(client.options)
        # If we have an active span, return baggage from there
        if span_streaming and self.streamed_span is not None:
            return self.streamed_span._to_baggage()
        elif not span_streaming and self.span is not None:
            return self.span._to_baggage()

        # else return baggage from the propagation context
        return self.get_active_propagation_context().get_baggage()

    def get_trace_context(self) -> "Dict[str, Any]":
        """
        Returns the Sentry "trace" context from the Propagation Context.
        """
        if (
            has_tracing_enabled(self.get_client().options)
            and self._span is not None
            and not isinstance(self._span, NoOpSpan)
        ):
            return self._span._get_trace_context()

        # if we are tracing externally (otel), those values take precedence
        external_propagation_context = get_external_propagation_context()
        if external_propagation_context:
            trace_id, span_id = external_propagation_context
            return {"trace_id": trace_id, "span_id": span_id}

        propagation_context = self.get_active_propagation_context()

        return {
            "trace_id": propagation_context.trace_id,
            "span_id": propagation_context.span_id,
            "parent_span_id": propagation_context.parent_span_id,
            "dynamic_sampling_context": propagation_context.dynamic_sampling_context,
        }

    def trace_propagation_meta(self, *args: "Any", **kwargs: "Any") -> str:
        """
        Return meta tags which should be injected into HTML templates
        to allow propagation of trace information.
        """
        span = kwargs.pop("span", None)
        if span is not None:
            logger.warning(
                "The parameter `span` in trace_propagation_meta() is deprecated and will be removed in the future."
            )

        meta = ""

        for name, content in self.iter_trace_propagation_headers():
            meta += f'<meta name="{name}" content="{content}">'

        return meta

    def iter_headers(self) -> "Iterator[Tuple[str, str]]":
        """
        Creates a generator which returns the `sentry-trace` and `baggage` headers from the Propagation Context.
        Deprecated: use PropagationContext.iter_headers instead.
        """
        if self._propagation_context is not None:
            yield from self._propagation_context.iter_headers()

    def iter_trace_propagation_headers(
        self, *args: "Any", **kwargs: "Any"
    ) -> "Generator[Tuple[str, str], None, None]":
        """
        Return HTTP headers which allow propagation of trace data.

        If a span is given, the trace data will taken from the span.
        If no span is given, the trace data is taken from the scope.
        """
        client = self.get_client()
        if not client.options.get("propagate_traces"):
            warnings.warn(
                "The `propagate_traces` parameter is deprecated. Please use `trace_propagation_targets` instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            return

        span = kwargs.pop("span", None)
        if not span:
            span_streaming = has_span_streaming_enabled(client.options)
            span = self.streamed_span if span_streaming else self.span

        if (
            has_tracing_enabled(client.options)
            and span is not None
            and not isinstance(span, NoOpSpan)
        ):
            for header in span._iter_headers():
                yield header
        elif has_external_propagation_context():
            # when we have an external_propagation_context (otlp)
            # we leave outgoing propagation to the propagator
            return
        else:
            for header in self.get_active_propagation_context().iter_headers():
                yield header

    def get_active_propagation_context(self) -> "PropagationContext":
        if self._propagation_context is not None:
            return self._propagation_context

        current_scope = self.get_current_scope()
        if current_scope._propagation_context is not None:
            return current_scope._propagation_context

        isolation_scope = self.get_isolation_scope()
        # should actually never happen, but just in case someone calls scope.clear
        if isolation_scope._propagation_context is None:
            isolation_scope._propagation_context = PropagationContext()
        return isolation_scope._propagation_context

    @classmethod
    def set_custom_sampling_context(
        cls, custom_sampling_context: "dict[str, Any]"
    ) -> None:
        cls.get_current_scope().get_active_propagation_context()._set_custom_sampling_context(
            custom_sampling_context
        )

    def clear(self) -> None:
        """Clears the entire scope."""
        self._level: "Optional[LogLevelStr]" = None
        self._fingerprint: "Optional[List[str]]" = None
        self._transaction: "Optional[str]" = None
        self._transaction_info: "dict[str, str]" = {}
        self._user: "Optional[Dict[str, Any]]" = None

        self._tags: "Dict[str, Any]" = {}
        self._contexts: "Dict[str, Dict[str, Any]]" = {}
        self._extras: "dict[str, Any]" = {}
        self._attachments: "List[Attachment]" = []

        self.clear_breadcrumbs()
        self._should_capture: bool = True

        self._span: "Optional[Union[Span, StreamedSpan]]" = None
        self._session: "Optional[Session]" = None
        self._force_auto_session_tracking: "Optional[bool]" = None

        self._profile: "Optional[Profile]" = None

        self._propagation_context = None

        # self._last_event_id is only applicable to isolation scopes
        self._last_event_id: "Optional[str]" = None
        self._flags: "Optional[FlagBuffer]" = None

        self._attributes: "Attributes" = {}

        self._gen_ai_conversation_id: "Optional[str]" = None

    @_attr_setter
    def level(self, value: "LogLevelStr") -> None:
        """
        When set this overrides the level.

        .. deprecated:: 1.0.0
            Use :func:`set_level` instead.

        :param value: The level to set.
        """
        logger.warning(
            "Deprecated: use .set_level() instead. This will be removed in the future."
        )

        self._level = value

    def set_level(self, value: "LogLevelStr") -> None:
        """
        Sets the level for the scope.

        :param value: The level to set.
        """
        self._level = value

    @_attr_setter
    def fingerprint(self, value: "Optional[List[str]]") -> None:
        """When set this overrides the default fingerprint."""
        self._fingerprint = value

    @property
    def transaction(self) -> "Any":
        # would be type: () -> Optional[Transaction], see https://github.com/python/mypy/issues/3004
        """Return the transaction (root span) in the scope, if any."""

        # there is no span/transaction on the scope
        if self._span is None:
            return None

        if isinstance(self._span, StreamedSpan):
            warnings.warn(
                "Scope.transaction is not available in streaming mode.",
                DeprecationWarning,
                stacklevel=2,
            )
            return None

        # there is an orphan span on the scope
        if self._span.containing_transaction is None:
            return None

        # there is either a transaction (which is its own containing
        # transaction) or a non-orphan span on the scope
        return self._span.containing_transaction

    @transaction.setter
    def transaction(self, value: "Any") -> None:
        # would be type: (Optional[str]) -> None, see https://github.com/python/mypy/issues/3004
        """When set this forces a specific transaction name to be set.

        Deprecated: use set_transaction_name instead."""

        # XXX: the docstring above is misleading. The implementation of
        # apply_to_event prefers an existing value of event.transaction over
        # anything set in the scope.
        # XXX: note that with the introduction of the Scope.transaction getter,
        # there is a semantic and type mismatch between getter and setter. The
        # getter returns a Transaction, the setter sets a transaction name.
        # Without breaking version compatibility, we could make the setter set a
        # transaction name or transaction (self._span) depending on the type of
        # the value argument.

        logger.warning(
            "Assigning to scope.transaction directly is deprecated: use scope.set_transaction_name() instead."
        )
        self._transaction = value
        if self._span:
            if isinstance(self._span, StreamedSpan):
                warnings.warn(
                    "Scope.transaction is not available in streaming mode.",
                    DeprecationWarning,
                    stacklevel=2,
                )
                return None

            if self._span.containing_transaction:
                self._span.containing_transaction.name = value

    def set_transaction_name(self, name: str, source: "Optional[str]" = None) -> None:
        """Set the transaction name and optionally the transaction source."""
        self._transaction = name
        if self._span:
            if isinstance(self._span, NoOpStreamedSpan):
                return

            elif isinstance(self._span, StreamedSpan):
                self._span._segment.name = name
                if source:
                    self._span._segment.set_attribute(
                        "sentry.segment.name.source", getattr(source, "value", source)
                    )

            elif self._span.containing_transaction:
                self._span.containing_transaction.name = name
                if source:
                    self._span.containing_transaction.source = source

        if source:
            self._transaction_info["source"] = source

    @_attr_setter
    def user(self, value: "Optional[Dict[str, Any]]") -> None:
        """When set a specific user is bound to the scope. Deprecated in favor of set_user."""
        warnings.warn(
            "The `Scope.user` setter is deprecated in favor of `Scope.set_user()`.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.set_user(value)

    def set_user(self, value: "Optional[Dict[str, Any]]") -> None:
        """Sets a user for the scope."""
        self._user = value

        session = self.get_isolation_scope()._session
        if session is not None:
            session.update(user=value)

    @property
    def span(self) -> "Optional[Span]":
        """Get/set current tracing span or transaction."""
        return self._span if isinstance(self._span, Span) else None

    @span.setter
    def span(self, span: "Optional[Span]") -> None:
        self._span = span
        # XXX: this differs from the implementation in JS, there Scope.setSpan
        # does not set Scope._transactionName.
        if isinstance(span, Transaction):
            transaction = span
            if transaction.name:
                self._transaction = transaction.name
                if transaction.source:
                    self._transaction_info["source"] = transaction.source

    @property
    def streamed_span(self) -> "Optional[StreamedSpan]":
        """Get/set current tracing span."""
        return self._span if isinstance(self._span, StreamedSpan) else None

    @streamed_span.setter
    def streamed_span(self, span: "Optional[StreamedSpan]") -> None:
        self._span = span

        # Also set _transaction and _transaction_info in streaming mode as this
        # is used for populating events and linking them to segments
        if type(span) is StreamedSpan and span._is_segment():
            self._transaction = span.name
            if span._attributes.get("sentry.segment.name.source"):
                self._transaction_info["source"] = str(
                    span._attributes["sentry.segment.name.source"]
                )

    @property
    def profile(self) -> "Optional[Profile]":
        return self._profile

    @profile.setter
    def profile(self, profile: "Optional[Profile]") -> None:
        self._profile = profile

    def set_tag(self, key: str, value: "Any") -> None:
        """
        Sets a tag for a key to a specific value.

        :param key: Key of the tag to set.

        :param value: Value of the tag to set.
        """
        self._tags[key] = value

    def set_tags(self, tags: "Mapping[str, object]") -> None:
        """Sets multiple tags at once.

        This method updates multiple tags at once. The tags are passed as a dictionary
        or other mapping type.

        Calling this method is equivalent to calling `set_tag` on each key-value pair
        in the mapping. If a tag key already exists in the scope, its value will be
        updated. If the tag key does not exist in the scope, the key-value pair will
        be added to the scope.

        This method only modifies tag keys in the `tags` mapping passed to the method.
        `scope.set_tags({})` is, therefore, a no-op.

        :param tags: A mapping of tag keys to tag values to set.
        """
        self._tags.update(tags)

    def remove_tag(self, key: str) -> None:
        """
        Removes a specific tag.

        :param key: Key of the tag to remove.
        """
        self._tags.pop(key, None)

    def set_context(
        self,
        key: str,
        value: "Dict[str, Any]",
    ) -> None:
        """
        Binds a context at a certain key to a specific value.
        """
        self._contexts[key] = value

    def remove_context(
        self,
        key: str,
    ) -> None:
        """Removes a context."""
        self._contexts.pop(key, None)

    def set_extra(
        self,
        key: str,
        value: "Any",
    ) -> None:
        """Sets an extra key to a specific value."""
        self._extras[key] = value

    def remove_extra(
        self,
        key: str,
    ) -> None:
        """Removes a specific extra key."""
        self._extras.pop(key, None)

    def set_conversation_id(self, conversation_id: str) -> None:
        """
        Sets the conversation ID for gen_ai spans.

        :param conversation_id: The conversation ID to set.
        """
        self._gen_ai_conversation_id = conversation_id

    def get_conversation_id(self) -> "Optional[str]":
        """
        Gets the conversation ID for gen_ai spans.

        :returns: The conversation ID, or None if not set.
        """
        return self._gen_ai_conversation_id

    def remove_conversation_id(self) -> None:
        """Removes the conversation ID."""
        self._gen_ai_conversation_id = None

    def clear_breadcrumbs(self) -> None:
        """Clears breadcrumb buffer."""
        self._breadcrumbs: "Deque[Breadcrumb]" = deque()
        self._n_breadcrumbs_truncated = 0

    def add_attachment(
        self,
        bytes: "Union[None, bytes, Callable[[], bytes]]" = None,
        filename: "Optional[str]" = None,
        path: "Optional[str]" = None,
        content_type: "Optional[str]" = None,
        add_to_transactions: bool = False,
    ) -> None:
        """Adds an attachment to future events sent from this scope.

        The parameters are the same as for the :py:class:`sentry_sdk.attachments.Attachment` constructor.
        """
        self._attachments.append(
            Attachment(
                bytes=bytes,
                path=path,
                filename=filename,
                content_type=content_type,
                add_to_transactions=add_to_transactions,
            )
        )

    def add_breadcrumb(
        self,
        crumb: "Optional[Breadcrumb]" = None,
        hint: "Optional[BreadcrumbHint]" = None,
        **kwargs: "Any",
    ) -> None:
        """
        Adds a breadcrumb.

        :param crumb: Dictionary with the data as the sentry v7/v8 protocol expects.

        :param hint: An optional value that can be used by `before_breadcrumb`
            to customize the breadcrumbs that are emitted.
        """
        client = self.get_client()

        if not client.is_active():
            logger.info("Dropped breadcrumb because no client bound")
            return

        before_breadcrumb = client.options.get("before_breadcrumb")
        max_breadcrumbs = client.options.get("max_breadcrumbs", DEFAULT_MAX_BREADCRUMBS)

        crumb = dict(crumb or ())
        crumb.update(kwargs)
        if not crumb:
            return

        hint = dict(hint or ())

        if crumb.get("timestamp") is None:
            crumb["timestamp"] = datetime.now(timezone.utc)
        if crumb.get("type") is None:
            crumb["type"] = "default"

        if before_breadcrumb is not None:
            new_crumb = before_breadcrumb(crumb, hint)
        else:
            new_crumb = crumb

        if new_crumb is not None:
            self._breadcrumbs.append(new_crumb)
        else:
            logger.info("before breadcrumb dropped breadcrumb (%s)", crumb)

        while len(self._breadcrumbs) > max_breadcrumbs:
            self._breadcrumbs.popleft()
            self._n_breadcrumbs_truncated += 1

    def start_transaction(
        self,
        transaction: "Optional[Transaction]" = None,
        instrumenter: str = INSTRUMENTER.SENTRY,
        custom_sampling_context: "Optional[SamplingContext]" = None,
        **kwargs: "Unpack[TransactionKwargs]",
    ) -> "Union[Transaction, NoOpSpan]":
        """
        Start and return a transaction.

        Start an existing transaction if given, otherwise create and start a new
        transaction with kwargs.

        This is the entry point to manual tracing instrumentation.

        A tree structure can be built by adding child spans to the transaction,
        and child spans to other spans. To start a new child span within the
        transaction or any span, call the respective `.start_child()` method.

        Every child span must be finished before the transaction is finished,
        otherwise the unfinished spans are discarded.

        When used as context managers, spans and transactions are automatically
        finished at the end of the `with` block. If not using context managers,
        call the `.finish()` method.

        When the transaction is finished, it will be sent to Sentry with all its
        finished child spans.

        :param transaction: The transaction to start. If omitted, we create and
            start a new transaction.
        :param instrumenter: This parameter is meant for internal use only. It
            will be removed in the next major version.
        :param custom_sampling_context: The transaction's custom sampling context.
        :param kwargs: Optional keyword arguments to be passed to the Transaction
            constructor. See :py:class:`sentry_sdk.tracing.Transaction` for
            available arguments.
        """
        client = self.get_client()
        if has_span_streaming_enabled(client.options):
            warnings.warn(
                "Scope.start_transaction is not available in streaming mode.",
                DeprecationWarning,
                stacklevel=2,
            )
            return NoOpSpan()

        kwargs.setdefault("scope", self)

        configuration_instrumenter = client.options["instrumenter"]

        if instrumenter != configuration_instrumenter:
            return NoOpSpan()

        try_autostart_continuous_profiler()

        custom_sampling_context = custom_sampling_context or {}

        # kwargs at this point has type TransactionKwargs, since we have removed
        # the client and custom_sampling_context from it.
        transaction_kwargs: "TransactionKwargs" = kwargs

        # if we haven't been given a transaction, make one
        if transaction is None:
            transaction = Transaction(**transaction_kwargs)

        # use traces_sample_rate, traces_sampler, and/or inheritance to make a
        # sampling decision
        sampling_context = {
            "transaction_context": transaction.to_json(),
            "parent_sampled": transaction.parent_sampled,
        }
        sampling_context.update(custom_sampling_context)
        transaction._set_initial_sampling_decision(sampling_context=sampling_context)

        # update the sample rate in the dsc
        if transaction.sample_rate is not None:
            propagation_context = self.get_active_propagation_context()
            baggage = propagation_context.baggage

            if baggage is not None:
                baggage.sentry_items["sample_rate"] = str(transaction.sample_rate)

            if transaction._baggage:
                transaction._baggage.sentry_items["sample_rate"] = str(
                    transaction.sample_rate
                )

        if transaction.sampled:
            profile = Profile(
                transaction.sampled, transaction._start_timestamp_monotonic_ns
            )
            profile._set_initial_sampling_decision(sampling_context=sampling_context)

            transaction._profile = profile

            transaction._continuous_profile = try_profile_lifecycle_trace_start()

            # Typically, the profiler is set when the transaction is created. But when
            # using the auto lifecycle, the profiler isn't running when the first
            # transaction is started. So make sure we update the profiler id on it.
            if transaction._continuous_profile is not None:
                transaction.set_profiler_id(get_profiler_id())

            # we don't bother to keep spans if we already know we're not going to
            # send the transaction
            max_spans = (client.options["_experiments"].get("max_spans")) or 1000
            transaction.init_span_recorder(maxlen=max_spans)

        return transaction

    def start_span(
        self, instrumenter: str = INSTRUMENTER.SENTRY, **kwargs: "Any"
    ) -> "Span":
        """
        Start a span whose parent is the currently active span or transaction, if any.

        The return value is a :py:class:`sentry_sdk.tracing.Span` instance,
        typically used as a context manager to start and stop timing in a `with`
        block.

        Only spans contained in a transaction are sent to Sentry. Most
        integrations start a transaction at the appropriate time, for example
        for every incoming HTTP request. Use
        :py:meth:`sentry_sdk.start_transaction` to start a new transaction when
        one is not already in progress.

        For supported `**kwargs` see :py:class:`sentry_sdk.tracing.Span`.

        The instrumenter parameter is deprecated for user code, and it will
        be removed in the next major version. Going forward, it should only
        be used by the SDK itself.
        """
        client = sentry_sdk.get_client()
        if has_span_streaming_enabled(client.options):
            warnings.warn(
                "Scope.start_span is not available in streaming mode.",
                DeprecationWarning,
                stacklevel=2,
            )
            return NoOpSpan()

        if kwargs.get("description") is not None:
            warnings.warn(
                "The `description` parameter is deprecated. Please use `name` instead.",
                DeprecationWarning,
                stacklevel=2,
            )

        with new_scope():
            kwargs.setdefault("scope", self)

            client = self.get_client()

            configuration_instrumenter = client.options["instrumenter"]

            if instrumenter != configuration_instrumenter:
                return NoOpSpan()

            # get current span or transaction
            span = self.span or self.get_isolation_scope().span
            if isinstance(span, StreamedSpan):
                # make mypy happy
                return NoOpSpan()

            if span is None:
                # New spans get the `trace_id` from the scope
                if "trace_id" not in kwargs:
                    propagation_context = self.get_active_propagation_context()
                    kwargs["trace_id"] = propagation_context.trace_id

                span = Span(**kwargs)
            else:
                # Children take `trace_id`` from the parent span.
                span = span.start_child(**kwargs)

            return span

    def start_streamed_span(
        self,
        name: str,
        attributes: "Optional[Attributes]",
        parent_span: "Optional[StreamedSpan]",
        active: bool,
    ) -> "StreamedSpan":
        # TODO: rename to start_span once we drop the old API
        if isinstance(parent_span, NoOpStreamedSpan):
            # parent_span is only set if the user explicitly set it
            logger.debug(
                "Ignored parent span provided. Span will be parented to the "
                "currently active span instead."
            )

        if parent_span is _DEFAULT_PARENT_SPAN or isinstance(
            parent_span, NoOpStreamedSpan
        ):
            parent_span = self.streamed_span

        # If no eligible parent_span was provided and there is no currently
        # active span, this is a new segment
        if parent_span is None:
            propagation_context = self.get_active_propagation_context()

            if is_ignored_span(name, attributes):
                return NoOpStreamedSpan(
                    scope=self,
                    segment=None,
                    trace_id=propagation_context.trace_id,
                    parent_span_id=propagation_context.parent_span_id,
                    parent_sampled=propagation_context.parent_sampled,
                    baggage=propagation_context.baggage,
                    unsampled_reason="ignored",
                )

            sampled, sample_rate, sample_rand, outcome = _make_sampling_decision(
                name,
                attributes,
                self,
            )

            if sample_rate is not None:
                self._update_sample_rate(sample_rate)

            if sampled is False or sampled is None:
                return NoOpStreamedSpan(
                    scope=self,
                    segment=None,
                    trace_id=propagation_context.trace_id,
                    parent_span_id=propagation_context.parent_span_id,
                    parent_sampled=propagation_context.parent_sampled,
                    baggage=propagation_context.baggage,
                    sampled=sampled,
                    unsampled_reason=outcome,
                    sample_rand=sample_rand,
                    sample_rate=sample_rate,
                )

            return StreamedSpan(
                name=name,
                attributes=attributes,
                active=active,
                scope=self,
                segment=None,
                trace_id=propagation_context.trace_id,
                parent_span_id=propagation_context.parent_span_id,
                parent_sampled=propagation_context.parent_sampled,
                baggage=propagation_context.baggage,
                sample_rand=sample_rand,
                sample_rate=sample_rate,
            )

        # This is a child span; take propagation context from the parent span
        with new_scope():
            if is_ignored_span(name, attributes):
                return NoOpStreamedSpan(
                    segment=parent_span._segment,
                    trace_id=parent_span.trace_id,
                    parent_span_id=parent_span.span_id,
                    parent_sampled=parent_span.sampled,
                    unsampled_reason="ignored",
                )

            if isinstance(parent_span, NoOpStreamedSpan):
                return NoOpStreamedSpan(
                    segment=parent_span._segment,
                    trace_id=parent_span.trace_id,
                    parent_span_id=parent_span.span_id,
                    parent_sampled=parent_span.sampled,
                    unsampled_reason=parent_span._unsampled_reason,
                )

            return StreamedSpan(
                name=name,
                attributes=attributes,
                active=active,
                scope=self,
                segment=parent_span._segment,
                trace_id=parent_span.trace_id,
                parent_span_id=parent_span.span_id,
                parent_sampled=parent_span.sampled,
            )

    def _update_sample_rate(self, sample_rate: float) -> None:
        # If we had to adjust the sample rate when setting the sampling decision
        # for a span, it needs to be updated in the propagation context too
        propagation_context = self.get_active_propagation_context()
        baggage = propagation_context.baggage

        if baggage is not None and baggage.sentry_items.get("sample_rate"):
            baggage.sentry_items["sample_rate"] = str(sample_rate)

    def continue_trace(
        self,
        environ_or_headers: "Dict[str, Any]",
        op: "Optional[str]" = None,
        name: "Optional[str]" = None,
        source: "Optional[str]" = None,
        origin: str = "manual",
    ) -> "Transaction":
        """
        Sets the propagation context from environment or headers and returns a transaction.
        """
        self.generate_propagation_context(environ_or_headers)

        # generate_propagation_context ensures that the propagation_context is not None.
        propagation_context = cast(PropagationContext, self._propagation_context)

        optional_kwargs = {}
        if name:
            optional_kwargs["name"] = name
        if source:
            optional_kwargs["source"] = source

        return Transaction(
            op=op,
            origin=origin,
            baggage=propagation_context.baggage,
            parent_sampled=propagation_context.parent_sampled,
            trace_id=propagation_context.trace_id,
            parent_span_id=propagation_context.parent_span_id,
            same_process_as_parent=False,
            **optional_kwargs,
        )

    def capture_event(
        self,
        event: "Event",
        hint: "Optional[Hint]" = None,
        scope: "Optional[Scope]" = None,
        **scope_kwargs: "Any",
    ) -> "Optional[str]":
        """
        Captures an event.

        Merges given scope data and calls :py:meth:`sentry_sdk.client._Client.capture_event`.

        :param event: A ready-made event that can be directly sent to Sentry.

        :param hint: Contains metadata about the event that can be read from `before_send`, such as the original exception object or a HTTP request object.

        :param scope: An optional :py:class:`sentry_sdk.Scope` to apply to events.
            The `scope` and `scope_kwargs` parameters are mutually exclusive.

        :param scope_kwargs: Optional data to apply to event.
            For supported `**scope_kwargs` see :py:meth:`sentry_sdk.Scope.update_from_kwargs`.
            The `scope` and `scope_kwargs` parameters are mutually exclusive.

        :returns: An `event_id` if the SDK decided to send the event (see :py:meth:`sentry_sdk.client._Client.capture_event`).
        """
        if disable_capture_event.get(False):
            return None

        scope = self._merge_scopes(scope, scope_kwargs)

        event_id = self.get_client().capture_event(event=event, hint=hint, scope=scope)

        if event_id is not None and event.get("type") != "transaction":
            self.get_isolation_scope()._last_event_id = event_id

        return event_id

    def _capture_log(self, log: "Optional[Log]") -> None:
        if log is None:
            return

        client = self.get_client()
        if not has_logs_enabled(client.options):
            return

        merged_scope = self._merge_scopes()

        debug = client.options.get("debug", False)
        if debug:
            logger.debug(
                f"[Sentry Logs] [{log.get('severity_text')}] {log.get('body')}"
            )

        client._capture_log(log, scope=merged_scope)

    def _capture_metric(self, metric: "Optional[Metric]") -> None:
        if metric is None:
            return

        client = self.get_client()
        if not has_metrics_enabled(client.options):
            return

        merged_scope = self._merge_scopes()

        debug = client.options.get("debug", False)
        if debug:
            logger.debug(
                f"[Sentry Metrics] [{metric.get('type')}] {metric.get('name')}: {metric.get('value')}"
            )

        client._capture_metric(metric, scope=merged_scope)

    def _capture_span(self, span: "Optional[StreamedSpan]") -> None:
        if span is None:
            return

        client = self.get_client()
        if not has_span_streaming_enabled(client.options):
            return

        merged_scope = self._merge_scopes()
        client._capture_span(span, scope=merged_scope)

    def capture_message(
        self,
        message: str,
        level: "Optional[LogLevelStr]" = None,
        scope: "Optional[Scope]" = None,
        **scope_kwargs: "Any",
    ) -> "Optional[str]":
        """
        Captures a message.

        :param message: The string to send as the message.

        :param level: If no level is provided, the default level is `info`.

        :param scope: An optional :py:class:`sentry_sdk.Scope` to apply to events.
            The `scope` and `scope_kwargs` parameters are mutually exclusive.

        :param scope_kwargs: Optional data to apply to event.
            For supported `**scope_kwargs` see :py:meth:`sentry_sdk.Scope.update_from_kwargs`.
            The `scope` and `scope_kwargs` parameters are mutually exclusive.

        :returns: An `event_id` if the SDK decided to send the event (see :py:meth:`sentry_sdk.client._Client.capture_event`).
        """
        if disable_capture_event.get(False):
            return None

        if level is None:
            level = "info"

        event: "Event" = {
            "message": message,
            "level": level,
        }

        return self.capture_event(event, scope=scope, **scope_kwargs)

    def capture_exception(
        self,
        error: "Optional[Union[BaseException, ExcInfo]]" = None,
        scope: "Optional[Scope]" = None,
        **scope_kwargs: "Any",
    ) -> "Optional[str]":
        """Captures an exception.

        :param error: An exception to capture. If `None`, `sys.exc_info()` will be used.

        :param scope: An optional :py:class:`sentry_sdk.Scope` to apply to events.
            The `scope` and `scope_kwargs` parameters are mutually exclusive.

        :param scope_kwargs: Optional data to apply to event.
            For supported `**scope_kwargs` see :py:meth:`sentry_sdk.Scope.update_from_kwargs`.
            The `scope` and `scope_kwargs` parameters are mutually exclusive.

        :returns: An `event_id` if the SDK decided to send the event (see :py:meth:`sentry_sdk.client._Client.capture_event`).
        """
        if disable_capture_event.get(False):
            return None

        if error is not None:
            exc_info = exc_info_from_error(error)
        else:
            exc_info = sys.exc_info()

        event, hint = event_from_exception(
            exc_info, client_options=self.get_client().options
        )

        try:
            return self.capture_event(event, hint=hint, scope=scope, **scope_kwargs)
        except Exception:
            capture_internal_exception(sys.exc_info())

        return None

    def start_session(self, *args: "Any", **kwargs: "Any") -> None:
        """Starts a new session."""
        session_mode = kwargs.pop("session_mode", "application")

        self.end_session()

        client = self.get_client()
        self._session = Session(
            release=client.options.get("release"),
            environment=client.options.get("environment"),
            user=self._user,
            session_mode=session_mode,
        )

    def end_session(self, *args: "Any", **kwargs: "Any") -> None:
        """Ends the current session if there is one."""
        session = self._session
        self._session = None

        if session is not None:
            session.close()
            self.get_client().capture_session(session)

    def stop_auto_session_tracking(self, *args: "Any", **kwargs: "Any") -> None:
        """Stops automatic session tracking.

        This temporarily session tracking for the current scope when called.
        To resume session tracking call `resume_auto_session_tracking`.
        """
        self.end_session()
        self._force_auto_session_tracking = False

    def resume_auto_session_tracking(self) -> None:
        """Resumes automatic session tracking for the current scope if
        disabled earlier.  This requires that generally automatic session
        tracking is enabled.
        """
        self._force_auto_session_tracking = None

    def add_event_processor(
        self,
        func: "EventProcessor",
    ) -> None:
        """Register a scope local event processor on the scope.

        :param func: This function behaves like `before_send.`
        """
        if len(self._event_processors) > 20:
            logger.warning(
                "Too many event processors on scope! Clearing list to free up some memory: %r",
                self._event_processors,
            )
            del self._event_processors[:]

        self._event_processors.append(func)

    def add_error_processor(
        self,
        func: "ErrorProcessor",
        cls: "Optional[Type[BaseException]]" = None,
    ) -> None:
        """Register a scope local error processor on the scope.

        :param func: A callback that works similar to an event processor but is invoked with the original exception info triple as second argument.

        :param cls: Optionally, only process exceptions of this type.
        """
        if cls is not None:
            cls_ = cls  # For mypy.
            real_func = func

            def func(event: "Event", exc_info: "ExcInfo") -> "Optional[Event]":
                try:
                    is_inst = isinstance(exc_info[1], cls_)
                except Exception:
                    is_inst = False
                if is_inst:
                    return real_func(event, exc_info)
                return event

        self._error_processors.append(func)

    def _apply_level_to_event(
        self, event: "Event", hint: "Hint", options: "Optional[Dict[str, Any]]"
    ) -> None:
        if self._level is not None:
            event["level"] = self._level

    def _apply_breadcrumbs_to_event(
        self, event: "Event", hint: "Hint", options: "Optional[Dict[str, Any]]"
    ) -> None:
        event.setdefault("breadcrumbs", {})

        # This check is just for mypy -
        if not isinstance(event["breadcrumbs"], AnnotatedValue):
            event["breadcrumbs"].setdefault("values", [])
            event["breadcrumbs"]["values"].extend(self._breadcrumbs)

        # Attempt to sort timestamps
        try:
            if not isinstance(event["breadcrumbs"], AnnotatedValue):
                for crumb in event["breadcrumbs"]["values"]:
                    if isinstance(crumb["timestamp"], str):
                        crumb["timestamp"] = datetime_from_isoformat(crumb["timestamp"])

                event["breadcrumbs"]["values"].sort(
                    key=lambda crumb: crumb["timestamp"]
                )
        except Exception as err:
            logger.debug("Error when sorting breadcrumbs", exc_info=err)
            pass

    def _apply_user_to_event(
        self, event: "Event", hint: "Hint", options: "Optional[Dict[str, Any]]"
    ) -> None:
        if event.get("user") is None and self._user is not None:
            event["user"] = self._user

    def _apply_transaction_name_to_event(
        self, event: "Event", hint: "Hint", options: "Optional[Dict[str, Any]]"
    ) -> None:
        if event.get("transaction") is None and self._transaction is not None:
            event["transaction"] = self._transaction

    def _apply_transaction_info_to_event(
        self, event: "Event", hint: "Hint", options: "Optional[Dict[str, Any]]"
    ) -> None:
        if event.get("transaction_info") is None and self._transaction_info is not None:
            event["transaction_info"] = self._transaction_info

    def _apply_fingerprint_to_event(
        self, event: "Event", hint: "Hint", options: "Optional[Dict[str, Any]]"
    ) -> None:
        if event.get("fingerprint") is None and self._fingerprint is not None:
            event["fingerprint"] = self._fingerprint

    def _apply_extra_to_event(
        self, event: "Event", hint: "Hint", options: "Optional[Dict[str, Any]]"
    ) -> None:
        if self._extras:
            event.setdefault("extra", {}).update(self._extras)

    def _apply_tags_to_event(
        self, event: "Event", hint: "Hint", options: "Optional[Dict[str, Any]]"
    ) -> None:
        if self._tags:
            event.setdefault("tags", {}).update(self._tags)

    def _apply_contexts_to_event(
        self, event: "Event", hint: "Hint", options: "Optional[Dict[str, Any]]"
    ) -> None:
        if self._contexts:
            event.setdefault("contexts", {}).update(self._contexts)

        contexts = event.setdefault("contexts", {})

        # Add "trace" context
        if contexts.get("trace") is None:
            contexts["trace"] = self.get_trace_context()

    def _apply_flags_to_event(
        self, event: "Event", hint: "Hint", options: "Optional[Dict[str, Any]]"
    ) -> None:
        flags = self.flags.get()
        if len(flags) > 0:
            event.setdefault("contexts", {}).setdefault("flags", {}).update(
                {"values": flags}
            )

    def _apply_scope_attributes_to_telemetry(
        self, telemetry: "Union[Log, Metric, StreamedSpan]"
    ) -> None:
        # TODO: turn Logs, Metrics into actual classes
        if isinstance(telemetry, dict):
            attributes = telemetry["attributes"]
        else:
            attributes = telemetry._attributes

        for attribute, value in self._attributes.items():
            if attribute not in attributes:
                attributes[attribute] = value

    def _apply_user_attributes_to_telemetry(
        self, telemetry: "Union[Log, Metric, StreamedSpan]"
    ) -> None:
        if isinstance(telemetry, dict):
            attributes = telemetry["attributes"]
        else:
            attributes = telemetry._attributes

        if not should_send_default_pii() or self._user is None:
            return

        for attribute_name, user_attribute in (
            ("user.id", "id"),
            ("user.name", "username"),
            ("user.email", "email"),
            ("user.ip_address", "ip_address"),
        ):
            if (
                user_attribute in self._user
                and attribute_name not in attributes
                and self._user[user_attribute] is not None
            ):
                attributes[attribute_name] = self._user[user_attribute]

    def _drop(self, cause: "Any", ty: str) -> "Optional[Any]":
        logger.info("%s (%s) dropped event", ty, cause)
        return None

    def run_error_processors(self, event: "Event", hint: "Hint") -> "Optional[Event]":
        """
        Runs the error processors on the event and returns the modified event.
        """
        exc_info = hint.get("exc_info")
        if exc_info is not None:
            error_processors = chain(
                self.get_global_scope()._error_processors,
                self.get_isolation_scope()._error_processors,
                self.get_current_scope()._error_processors,
            )

            for error_processor in error_processors:
                new_event = error_processor(event, exc_info)
                if new_event is None:
                    return self._drop(error_processor, "error processor")

                event = new_event

        return event

    def run_event_processors(self, event: "Event", hint: "Hint") -> "Optional[Event]":
        """
        Runs the event processors on the event and returns the modified event.
        """
        ty = event.get("type")
        is_check_in = ty == "check_in"

        if not is_check_in:
            # Get scopes without creating them to prevent infinite recursion
            isolation_scope = _isolation_scope.get()
            current_scope = _current_scope.get()

            event_processors = chain(
                global_event_processors,
                _global_scope and _global_scope._event_processors or [],
                isolation_scope and isolation_scope._event_processors or [],
                current_scope and current_scope._event_processors or [],
            )

            for event_processor in event_processors:
                new_event = event
                with capture_internal_exceptions():
                    new_event = event_processor(event, hint)
                if new_event is None:
                    return self._drop(event_processor, "event processor")
                event = new_event

        return event

    @_disable_capture
    def apply_to_event(
        self,
        event: "Event",
        hint: "Hint",
        options: "Optional[Dict[str, Any]]" = None,
    ) -> "Optional[Event]":
        """Applies the information contained on the scope to the given event."""
        ty = event.get("type")
        is_transaction = ty == "transaction"
        is_check_in = ty == "check_in"

        # put all attachments into the hint. This lets callbacks play around
        # with attachments. We also later pull this out of the hint when we
        # create the envelope.
        attachments_to_send = hint.get("attachments") or []
        for attachment in self._attachments:
            if not is_transaction or attachment.add_to_transactions:
                attachments_to_send.append(attachment)
        hint["attachments"] = attachments_to_send

        self._apply_contexts_to_event(event, hint, options)

        if is_check_in:
            # Check-ins only support the trace context, strip all others
            event["contexts"] = {
                "trace": event.setdefault("contexts", {}).get("trace", {})
            }

        if not is_check_in:
            self._apply_level_to_event(event, hint, options)
            self._apply_fingerprint_to_event(event, hint, options)
            self._apply_user_to_event(event, hint, options)
            self._apply_transaction_name_to_event(event, hint, options)
            self._apply_transaction_info_to_event(event, hint, options)
            self._apply_tags_to_event(event, hint, options)
            self._apply_extra_to_event(event, hint, options)

        if not is_transaction and not is_check_in:
            self._apply_breadcrumbs_to_event(event, hint, options)
            self._apply_flags_to_event(event, hint, options)

        event = self.run_error_processors(event, hint)
        if event is None:
            return None

        event = self.run_event_processors(event, hint)
        if event is None:
            return None

        return event

    @_disable_capture
    def apply_to_telemetry(self, telemetry: "Union[Log, Metric, StreamedSpan]") -> None:
        # Attributes-based events and telemetry go through here (logs, metrics,
        # spansV2)
        if not isinstance(telemetry, StreamedSpan):
            trace_context = self.get_trace_context()
            trace_id = trace_context.get("trace_id")
            if telemetry.get("trace_id") is None and trace_id is not None:
                telemetry["trace_id"] = trace_id

            # span_id should only be populated if there's an active span. We can't
            # use the trace_context here because it synthesizes a span_id if there
            # isn't one
            if telemetry.get("span_id") is None:
                if self._span is not None and not isinstance(
                    self._span, (NoOpStreamedSpan, NoOpSpan)
                ):
                    telemetry["span_id"] = self._span.span_id
                else:
                    external_propagation_context = get_external_propagation_context()
                    if external_propagation_context:
                        _, span_id = external_propagation_context
                        if span_id is not None:
                            telemetry["span_id"] = span_id

        self._apply_scope_attributes_to_telemetry(telemetry)
        self._apply_user_attributes_to_telemetry(telemetry)

    def update_from_scope(self, scope: "Scope") -> None:
        """Update the scope with another scope's data."""
        if scope._level is not None:
            self._level = scope._level
        if scope._fingerprint is not None:
            self._fingerprint = scope._fingerprint
        if scope._transaction is not None:
            self._transaction = scope._transaction
        if scope._transaction_info is not None:
            self._transaction_info.update(scope._transaction_info)
        if scope._user is not None:
            self._user = scope._user
        if scope._tags:
            self._tags.update(scope._tags)
        if scope._contexts:
            self._contexts.update(scope._contexts)
        if scope._extras:
            self._extras.update(scope._extras)
        if scope._breadcrumbs:
            self._breadcrumbs.extend(scope._breadcrumbs)
        if scope._n_breadcrumbs_truncated:
            self._n_breadcrumbs_truncated = (
                self._n_breadcrumbs_truncated + scope._n_breadcrumbs_truncated
            )
        if scope._gen_ai_original_message_count:
            self._gen_ai_original_message_count.update(
                scope._gen_ai_original_message_count
            )
        if scope._gen_ai_conversation_id:
            self._gen_ai_conversation_id = scope._gen_ai_conversation_id
        if scope._span:
            self._span = scope._span
        if scope._attachments:
            self._attachments.extend(scope._attachments)
        if scope._profile:
            self._profile = scope._profile
        if scope._propagation_context:
            self._propagation_context = scope._propagation_context
        if scope._session:
            self._session = scope._session
        if scope._flags:
            if not self._flags:
                self._flags = deepcopy(scope._flags)
            else:
                for flag in scope._flags.get():
                    self._flags.set(flag["flag"], flag["result"])
        if scope._attributes:
            self._attributes.update(scope._attributes)

    def update_from_kwargs(
        self,
        user: "Optional[Any]" = None,
        level: "Optional[LogLevelStr]" = None,
        extras: "Optional[Dict[str, Any]]" = None,
        contexts: "Optional[Dict[str, Dict[str, Any]]]" = None,
        tags: "Optional[Dict[str, str]]" = None,
        fingerprint: "Optional[List[str]]" = None,
        attributes: "Optional[Attributes]" = None,
    ) -> None:
        """Update the scope's attributes."""
        if level is not None:
            self._level = level
        if user is not None:
            self._user = user
        if extras is not None:
            self._extras.update(extras)
        if contexts is not None:
            self._contexts.update(contexts)
        if tags is not None:
            self._tags.update(tags)
        if fingerprint is not None:
            self._fingerprint = fingerprint
        if attributes is not None:
            self._attributes.update(attributes)

    def __repr__(self) -> str:
        return "<%s id=%s name=%s type=%s>" % (
            self.__class__.__name__,
            hex(id(self)),
            self._name,
            self._type,
        )

    @property
    def flags(self) -> "FlagBuffer":
        if self._flags is None:
            max_flags = (
                self.get_client().options["_experiments"].get("max_flags")
                or DEFAULT_FLAG_CAPACITY
            )
            self._flags = FlagBuffer(capacity=max_flags)
        return self._flags

    def set_attribute(self, attribute: str, value: "AttributeValue") -> None:
        """
        Set an attribute on the scope.

        Any attributes-based telemetry (logs, metrics) captured while this scope
        is active will inherit attributes set on the scope.
        """
        self._attributes[attribute] = format_attribute(value)

    def remove_attribute(self, attribute: str) -> None:
        """Remove an attribute if set on the scope. No-op if there is no such attribute."""
        try:
            del self._attributes[attribute]
        except KeyError:
            pass


@contextmanager
def new_scope() -> "Generator[Scope, None, None]":
    """
    .. versionadded:: 2.0.0

    Context manager that forks the current scope and runs the wrapped code in it.
    After the wrapped code is executed, the original scope is restored.

    Example Usage:

    .. code-block:: python

        import sentry_sdk

        with sentry_sdk.new_scope() as scope:
            scope.set_tag("color", "green")
            sentry_sdk.capture_message("hello") # will include `color` tag.

        sentry_sdk.capture_message("hello, again") # will NOT include `color` tag.

    """
    # fork current scope
    current_scope = Scope.get_current_scope()
    new_scope = current_scope.fork()
    token = _current_scope.set(new_scope)

    try:
        yield new_scope

    finally:
        try:
            # restore original scope
            _current_scope.reset(token)
        except (LookupError, ValueError):
            capture_internal_exception(sys.exc_info())


@contextmanager
def use_scope(scope: "Scope") -> "Generator[Scope, None, None]":
    """
    .. versionadded:: 2.0.0

    Context manager that uses the given `scope` and runs the wrapped code in it.
    After the wrapped code is executed, the original scope is restored.

    Example Usage:
    Suppose the variable `scope` contains a `Scope` object, which is not currently
    the active scope.

    .. code-block:: python

        import sentry_sdk

        with sentry_sdk.use_scope(scope):
            scope.set_tag("color", "green")
            sentry_sdk.capture_message("hello") # will include `color` tag.

        sentry_sdk.capture_message("hello, again") # will NOT include `color` tag.

    """
    # set given scope as current scope
    token = _current_scope.set(scope)

    try:
        yield scope

    finally:
        try:
            # restore original scope
            _current_scope.reset(token)
        except (LookupError, ValueError):
            capture_internal_exception(sys.exc_info())


@contextmanager
def isolation_scope() -> "Generator[Scope, None, None]":
    """
    .. versionadded:: 2.0.0

    Context manager that forks the current isolation scope and runs the wrapped code in it.
    The current scope is also forked to not bleed data into the existing current scope.
    After the wrapped code is executed, the original scopes are restored.

    Example Usage:

    .. code-block:: python

        import sentry_sdk

        with sentry_sdk.isolation_scope() as scope:
            scope.set_tag("color", "green")
            sentry_sdk.capture_message("hello") # will include `color` tag.

        sentry_sdk.capture_message("hello, again") # will NOT include `color` tag.

    """
    # fork current scope
    current_scope = Scope.get_current_scope()
    forked_current_scope = current_scope.fork()
    current_token = _current_scope.set(forked_current_scope)

    # fork isolation scope
    isolation_scope = Scope.get_isolation_scope()
    new_isolation_scope = isolation_scope.fork()
    isolation_token = _isolation_scope.set(new_isolation_scope)

    try:
        yield new_isolation_scope

    finally:
        # restore original scopes
        try:
            _current_scope.reset(current_token)
        except (LookupError, ValueError):
            capture_internal_exception(sys.exc_info())

        try:
            _isolation_scope.reset(isolation_token)
        except (LookupError, ValueError):
            capture_internal_exception(sys.exc_info())


@contextmanager
def use_isolation_scope(isolation_scope: "Scope") -> "Generator[Scope, None, None]":
    """
    .. versionadded:: 2.0.0

    Context manager that uses the given `isolation_scope` and runs the wrapped code in it.
    The current scope is also forked to not bleed data into the existing current scope.
    After the wrapped code is executed, the original scopes are restored.

    Example Usage:

    .. code-block:: python

        import sentry_sdk

        with sentry_sdk.isolation_scope() as scope:
            scope.set_tag("color", "green")
            sentry_sdk.capture_message("hello") # will include `color` tag.

        sentry_sdk.capture_message("hello, again") # will NOT include `color` tag.

    """
    # fork current scope
    current_scope = Scope.get_current_scope()
    forked_current_scope = current_scope.fork()
    current_token = _current_scope.set(forked_current_scope)

    # set given scope as isolation scope
    isolation_token = _isolation_scope.set(isolation_scope)

    try:
        yield isolation_scope

    finally:
        # restore original scopes
        try:
            _current_scope.reset(current_token)
        except (LookupError, ValueError):
            capture_internal_exception(sys.exc_info())

        try:
            _isolation_scope.reset(isolation_token)
        except (LookupError, ValueError):
            capture_internal_exception(sys.exc_info())


def should_send_default_pii() -> bool:
    """Shortcut for `Scope.get_client().should_send_default_pii()`."""
    return Scope.get_client().should_send_default_pii()


# Circular imports
from sentry_sdk.client import NonRecordingClient

if TYPE_CHECKING:
    import sentry_sdk.client



"""
Separate module for tests that check backwards compatibility of the Hub API with 1.x.
These tests should be removed once we remove the Hub API, likely in the next major.

All tests in this module are run with hub isolation, provided by `isolate_hub` autouse
fixture, defined in `conftest.py`.
"""



from __future__ import annotations

import functools
from typing import Any, Dict

import pytest
from starlite import AbstractMiddleware, Controller, LoggingConfig, Starlite, get
from starlite.middleware import LoggingMiddlewareConfig, RateLimitConfig
from starlite.middleware.session.memory_backend import MemoryBackendConfig
from starlite.testing import TestClient

import sentry_sdk
from sentry_sdk import capture_message
from sentry_sdk.integrations.starlite import StarliteIntegration


def starlite_app_factory(middleware=None, debug=True, exception_handlers=None):
    class MyController(Controller):
        path = "/controller"

        @get("/error")
        async def controller_error(self) -> None:
            raise Exception("Whoa")

    @get("/some_url")
    async def homepage_handler() -> "Dict[str, Any]":
        1 / 0
        return {"status": "ok"}

    @get("/custom_error", name="custom_name")
    async def custom_error() -> Any:
        raise Exception("Too Hot")

    @get("/message")
    async def message() -> "Dict[str, Any]":
        capture_message("hi")
        return {"status": "ok"}

    @get("/message/{message_id:str}")
    async def message_with_id() -> "Dict[str, Any]":
        capture_message("hi")
        return {"status": "ok"}

    logging_config = LoggingConfig()

    app = Starlite(
        route_handlers=[
            homepage_handler,
            custom_error,
            message,
            message_with_id,
            MyController,
        ],
        debug=debug,
        middleware=middleware,
        logging_config=logging_config,
        exception_handlers=exception_handlers,
    )

    return app


@pytest.mark.parametrize(
    "test_url,expected_error,expected_message,expected_tx_name",
    [
        (
            "/some_url",
            ZeroDivisionError,
            "division by zero",
            "tests.integrations.starlite.test_starlite.starlite_app_factory.<locals>.homepage_handler",
        ),
        (
            "/custom_error",
            Exception,
            "Too Hot",
            "custom_name",
        ),
        (
            "/controller/error",
            Exception,
            "Whoa",
            "partial(<function tests.integrations.starlite.test_starlite.starlite_app_factory.<locals>.MyController.controller_error>)",
        ),
    ],
)
def test_catch_exceptions(
    sentry_init,
    capture_exceptions,
    capture_events,
    test_url,
    expected_error,
    expected_message,
    expected_tx_name,
):
    sentry_init(integrations=[StarliteIntegration()])
    starlite_app = starlite_app_factory()
    exceptions = capture_exceptions()
    events = capture_events()

    client = TestClient(starlite_app)
    try:
        client.get(test_url)
    except Exception:
        pass

    (exc,) = exceptions
    assert isinstance(exc, expected_error)
    assert str(exc) == expected_message

    (event,) = events
    assert event["transaction"] == expected_tx_name
    assert event["exception"]["values"][0]["mechanism"]["type"] == "starlite"


@pytest.mark.parametrize(
    "test_url,expected_tx_name",
    [
        (
            "/some_url",
            "tests.integrations.starlite.test_starlite.starlite_app_factory.<locals>.homepage_handler",
        ),
        (
            "/custom_error",
            "custom_name",
        ),
        (
            "/controller/error",
            "partial(<function tests.integrations.starlite.test_starlite.starlite_app_factory.<locals>.MyController.controller_error>)",
        ),
    ],
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_transaction_name_and_source(
    sentry_init,
    capture_events,
    test_url,
    expected_tx_name,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[StarliteIntegration()],
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )
    starlite_app = starlite_app_factory()
    client = TestClient(starlite_app)

    if span_streaming:
        items = capture_items("span")

        try:
            client.get(test_url)
        except Exception:
            pass

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        spans = [span for span in spans if expected_tx_name in span["name"]]
        assert len(spans) == 1
        assert spans[0]["attributes"]["sentry.segment.name.source"] == "component"
    else:
        events = capture_events()

        try:
            client.get(test_url)
        except Exception:
            pass

        (_, transaction) = events
        assert expected_tx_name in transaction["transaction"]
        assert transaction["transaction_info"] == {"source": "component"}


@pytest.mark.parametrize("span_streaming", [True, False])
def test_middleware_spans(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[StarliteIntegration()],
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    logging_config = LoggingMiddlewareConfig()
    session_config = MemoryBackendConfig()
    rate_limit_config = RateLimitConfig(rate_limit=("hour", 5))

    starlite_app = starlite_app_factory(
        middleware=[
            session_config.middleware,
            logging_config.middleware,
            rate_limit_config.middleware,
        ]
    )

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    client = TestClient(
        starlite_app, raise_server_exceptions=False, base_url="http://testserver.local"
    )
    client.get("/message")

    expected = {"SessionMiddleware", "LoggingMiddleware", "RateLimitMiddleware"}

    if span_streaming:
        sentry_sdk.flush()

        middleware_spans = [
            item.payload
            for item in items
            if item.payload.get("attributes", {}).get("sentry.op")
            == "middleware.starlite"
        ]
        assert len(middleware_spans) == 3

        found = set()
        for span in middleware_spans:
            assert span["name"] in expected
            assert span["name"] not in found
            found.add(span["name"])
            assert span["name"] == span["attributes"]["middleware.name"]
    else:
        (_, transaction_event) = events

        found = set()
        middleware_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == "middleware.starlite"
        ]
        assert len(middleware_spans) == 3

        for span in middleware_spans:
            assert span["description"] in expected
            assert span["description"] not in found
            found.add(span["description"])
            assert span["description"] == span["tags"]["starlite.middleware_name"]


@pytest.mark.parametrize("span_streaming", [True, False])
def test_middleware_callback_spans(
    sentry_init, capture_events, capture_items, span_streaming
):
    class SampleMiddleware(AbstractMiddleware):
        async def __call__(self, scope, receive, send) -> None:
            async def do_stuff(message):
                if message["type"] == "http.response.start":
                    # do something here.
                    pass
                await send(message)

            await self.app(scope, receive, do_stuff)

    sentry_init(
        traces_sample_rate=1.0,
        integrations=[StarliteIntegration()],
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )
    starlite_app = starlite_app_factory(middleware=[SampleMiddleware])

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    client = TestClient(starlite_app, raise_server_exceptions=False)
    client.get("/message")

    expected_starlite_spans = [
        {
            "op": "middleware.starlite",
            "description": "SampleMiddleware",
            "tags": {"starlite.middleware_name": "SampleMiddleware"},
        },
        {
            "op": "middleware.starlite.send",
            "description": "SentryAsgiMiddleware._run_app.<locals>._sentry_wrapped_send",
            "tags": {"starlite.middleware_name": "SampleMiddleware"},
        },
        {
            "op": "middleware.starlite.send",
            "description": "SentryAsgiMiddleware._run_app.<locals>._sentry_wrapped_send",
            "tags": {"starlite.middleware_name": "SampleMiddleware"},
        },
    ]

    if span_streaming:
        sentry_sdk.flush()

        actual_starlite_spans = [
            item.payload
            for item in items
            if "middleware.starlite"
            in item.payload.get("attributes", {}).get("sentry.op", "")
        ]
        assert len(actual_starlite_spans) == 3

        def is_matching_span_streaming(expected_span, actual_span):
            return (
                expected_span["op"] == actual_span["attributes"]["sentry.op"]
                and expected_span["description"] == actual_span["name"]
                and expected_span["tags"]["starlite.middleware_name"]
                == actual_span["attributes"]["middleware.name"]
            )

        for expected_span in expected_starlite_spans:
            assert any(
                is_matching_span_streaming(expected_span, actual_span)
                for actual_span in actual_starlite_spans
            )
    else:
        (_, transaction_events) = events

        def is_matching_span(expected_span, actual_span):
            return (
                expected_span["op"] == actual_span["op"]
                and expected_span["description"] == actual_span["description"]
                and expected_span["tags"] == actual_span["tags"]
            )

        actual_starlite_spans = list(
            span
            for span in transaction_events["spans"]
            if "middleware.starlite" in span["op"]
        )
        assert len(actual_starlite_spans) == 3

        for expected_span in expected_starlite_spans:
            assert any(
                is_matching_span(expected_span, actual_span)
                for actual_span in actual_starlite_spans
            )


def test_middleware_receive_send(sentry_init, capture_events):
    class SampleReceiveSendMiddleware(AbstractMiddleware):
        async def __call__(self, scope, receive, send):
            message = await receive()
            assert message
            assert message["type"] == "http.request"

            send_output = await send({"type": "something-unimportant"})
            assert send_output is None

            await self.app(scope, receive, send)

    sentry_init(
        traces_sample_rate=1.0,
        integrations=[StarliteIntegration()],
    )
    starlite_app = starlite_app_factory(middleware=[SampleReceiveSendMiddleware])

    client = TestClient(starlite_app, raise_server_exceptions=False)
    # See SampleReceiveSendMiddleware.__call__ above for assertions of correct behavior
    client.get("/message")


@pytest.mark.parametrize("span_streaming", [True, False])
def test_middleware_partial_receive_send(
    sentry_init, capture_events, capture_items, span_streaming
):
    class SamplePartialReceiveSendMiddleware(AbstractMiddleware):
        async def __call__(self, scope, receive, send):
            message = await receive()
            assert message
            assert message["type"] == "http.request"

            send_output = await send({"type": "something-unimportant"})
            assert send_output is None

            async def my_receive(*args, **kwargs):
                pass

            async def my_send(*args, **kwargs):
                pass

            partial_receive = functools.partial(my_receive)
            partial_send = functools.partial(my_send)

            await self.app(scope, partial_receive, partial_send)

    sentry_init(
        traces_sample_rate=1.0,
        integrations=[StarliteIntegration()],
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )
    starlite_app = starlite_app_factory(middleware=[SamplePartialReceiveSendMiddleware])

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    client = TestClient(starlite_app, raise_server_exceptions=False)
    # See SamplePartialReceiveSendMiddleware.__call__ above for assertions of correct behavior
    client.get("/message")

    expected_starlite_spans = [
        {
            "op": "middleware.starlite",
            "description": "SamplePartialReceiveSendMiddleware",
            "tags": {"starlite.middleware_name": "SamplePartialReceiveSendMiddleware"},
        },
        {
            "op": "middleware.starlite.receive",
            "description": "TestClientTransport.create_receive.<locals>.receive",
            "tags": {"starlite.middleware_name": "SamplePartialReceiveSendMiddleware"},
        },
        {
            "op": "middleware.starlite.send",
            "description": "SentryAsgiMiddleware._run_app.<locals>._sentry_wrapped_send",
            "tags": {"starlite.middleware_name": "SamplePartialReceiveSendMiddleware"},
        },
    ]

    if span_streaming:
        sentry_sdk.flush()

        actual_starlite_spans = [
            item.payload
            for item in items
            if "middleware.starlite"
            in item.payload.get("attributes", {}).get("sentry.op", "")
        ]
        assert len(actual_starlite_spans) == 3

        def is_matching_span_streaming(expected_span, actual_span):
            return (
                expected_span["op"] == actual_span["attributes"]["sentry.op"]
                and actual_span["name"].startswith(expected_span["description"])
                and expected_span["tags"]["starlite.middleware_name"]
                == actual_span["attributes"]["middleware.name"]
            )

        for expected_span in expected_starlite_spans:
            assert any(
                is_matching_span_streaming(expected_span, actual_span)
                for actual_span in actual_starlite_spans
            )
    else:
        (_, transaction_events) = events

        def is_matching_span(expected_span, actual_span):
            return (
                expected_span["op"] == actual_span["op"]
                and actual_span["description"].startswith(expected_span["description"])
                and expected_span["tags"] == actual_span["tags"]
            )

        actual_starlite_spans = list(
            span
            for span in transaction_events["spans"]
            if "middleware.starlite" in span["op"]
        )
        assert len(actual_starlite_spans) == 3

        for expected_span in expected_starlite_spans:
            assert any(
                is_matching_span(expected_span, actual_span)
                for actual_span in actual_starlite_spans
            )


@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        integrations=[StarliteIntegration()],
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    logging_config = LoggingMiddlewareConfig()
    session_config = MemoryBackendConfig()
    rate_limit_config = RateLimitConfig(rate_limit=("hour", 5))

    starlite_app = starlite_app_factory(
        middleware=[
            session_config.middleware,
            logging_config.middleware,
            rate_limit_config.middleware,
        ]
    )

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    client = TestClient(
        starlite_app, raise_server_exceptions=False, base_url="http://testserver.local"
    )
    client.get("/message")

    if span_streaming:
        sentry_sdk.flush()

        starlite_items = [
            item
            for item in items
            if "starlite" in item.payload.get("attributes", {}).get("sentry.op", "")
        ]
        assert len(starlite_items) > 0
        for item in starlite_items:
            assert item.payload["attributes"]["sentry.origin"] == "auto.http.starlite"
    else:
        (_, event) = events

        assert event["contexts"]["trace"]["origin"] == "auto.http.starlite"
        starlite_spans = [span for span in event["spans"] if "starlite" in span["op"]]
        assert len(starlite_spans) > 0
        for span in starlite_spans:
            assert span["origin"] == "auto.http.starlite"


@pytest.mark.parametrize(
    "is_send_default_pii",
    [
        True,
        False,
    ],
    ids=[
        "send_default_pii=True",
        "send_default_pii=False",
    ],
)
def test_starlite_scope_user_on_exception_event(
    sentry_init, capture_exceptions, capture_events, is_send_default_pii
):
    class TestUserMiddleware(AbstractMiddleware):
        async def __call__(self, scope, receive, send):
            scope["user"] = {
                "email": "lennon@thebeatles.com",
                "username": "john",
                "id": "1",
            }
            await self.app(scope, receive, send)

    sentry_init(
        integrations=[StarliteIntegration()], send_default_pii=is_send_default_pii
    )
    starlite_app = starlite_app_factory(middleware=[TestUserMiddleware])
    exceptions = capture_exceptions()
    events = capture_events()

    # This request intentionally raises an exception
    client = TestClient(starlite_app)
    try:
        client.get("/some_url")
    except Exception:
        pass

    assert len(exceptions) == 1
    assert len(events) == 1
    (event,) = events

    if is_send_default_pii:
        assert "user" in event
        assert event["user"] == {
            "email": "lennon@thebeatles.com",
            "username": "john",
            "id": "1",
        }
    else:
        assert "user" not in event



import itertools
from enum import Enum
from typing import TYPE_CHECKING

DEFAULT_MAX_VALUE_LENGTH = None

DEFAULT_MAX_STACK_FRAMES = 100
DEFAULT_ADD_FULL_STACK = False


# Also needs to be at the top to prevent circular import
class EndpointType(Enum):
    """
    The type of an endpoint. This is an enum, rather than a constant, for historical reasons
    (the old /store endpoint). The enum also preserve future compatibility, in case we ever
    have a new endpoint.
    """

    ENVELOPE = "envelope"
    OTLP_TRACES = "integration/otlp/v1/traces"


class CompressionAlgo(Enum):
    GZIP = "gzip"
    BROTLI = "br"


if TYPE_CHECKING:
    from typing import (
        AbstractSet,
        Any,
        Callable,
        Dict,
        List,
        Optional,
        Sequence,
        Tuple,
        Type,
        Union,
    )

    from typing_extensions import Literal, TypedDict

    import sentry_sdk
    from sentry_sdk._types import (
        BreadcrumbProcessor,
        ContinuousProfilerMode,
        DataCollectionUserOptions,
        Event,
        EventProcessor,
        Hint,
        IgnoreSpansConfig,
        Log,
        Metric,
        ProfilerMode,
        SpanJSON,
        TracesSampler,
        TransactionProcessor,
    )

    # Experiments are feature flags to enable and disable certain unstable SDK
    # functionality. Changing them from the defaults (`None`) in production
    # code is highly discouraged. They are not subject to any stability
    # guarantees such as the ones from semantic versioning.
    Experiments = TypedDict(
        "Experiments",
        {
            "max_spans": Optional[int],
            "max_flags": Optional[int],
            "record_sql_params": Optional[bool],
            "continuous_profiling_auto_start": Optional[bool],
            "continuous_profiling_mode": Optional[ContinuousProfilerMode],
            "otel_powered_performance": Optional[bool],
            "transport_zlib_compression_level": Optional[int],
            "transport_compression_level": Optional[int],
            "transport_compression_algo": Optional[CompressionAlgo],
            "transport_num_pools": Optional[int],
            "transport_http2": Optional[bool],
            "transport_async": Optional[bool],
            "enable_logs": Optional[bool],
            "before_send_log": Optional[Callable[[Log, Hint], Optional[Log]]],
            "enable_metrics": Optional[bool],
            "before_send_metric": Optional[Callable[[Metric, Hint], Optional[Metric]]],
            "trace_lifecycle": Optional[Literal["static", "stream"]],
            "ignore_spans": Optional[IgnoreSpansConfig],
            "before_send_span": Optional[
                Callable[[SpanJSON, Hint], Optional[SpanJSON]]
            ],
            "suppress_asgi_chained_exceptions": Optional[bool],
            "data_collection": Optional[DataCollectionUserOptions],
        },
        total=False,
    )

DEFAULT_QUEUE_SIZE = 100
DEFAULT_MAX_BREADCRUMBS = 100
MATCH_ALL = r".*"

FALSE_VALUES = [
    "false",
    "no",
    "off",
    "n",
    "0",
]


class SPANTEMPLATE(str, Enum):
    DEFAULT = "default"
    AI_AGENT = "ai_agent"
    AI_TOOL = "ai_tool"
    AI_CHAT = "ai_chat"

    def __str__(self) -> str:
        return self.value


class INSTRUMENTER:
    SENTRY = "sentry"
    OTEL = "otel"


class SPANNAME:
    DB_COMMIT = "COMMIT"
    DB_ROLLBACK = "ROLLBACK"


class SPANDATA:
    """
    Additional information describing the type of the span.
    See: https://develop.sentry.dev/sdk/performance/span-data-conventions/
    """

    AI_CITATIONS = "ai.citations"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_* attributes instead.

    References or sources cited by the AI model in its response.
    Example: ["Smith et al. 2020", "Jones 2019"]
    """

    AI_DOCUMENTS = "ai.documents"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_* attributes instead.

    Documents or content chunks used as context for the AI model.
    Example: ["doc1.txt", "doc2.pdf"]
    """

    AI_FINISH_REASON = "ai.finish_reason"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_RESPONSE_FINISH_REASONS instead.

    The reason why the model stopped generating.
    Example: "length"
    """

    AI_FREQUENCY_PENALTY = "ai.frequency_penalty"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_REQUEST_FREQUENCY_PENALTY instead.

    Used to reduce repetitiveness of generated tokens.
    Example: 0.5
    """

    AI_FUNCTION_CALL = "ai.function_call"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_RESPONSE_TOOL_CALLS instead.

    For an AI model call, the function that was called. This is deprecated for OpenAI, and replaced by tool_calls
    """

    AI_GENERATION_ID = "ai.generation_id"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_RESPONSE_ID instead.

    Unique identifier for the completion.
    Example: "gen_123abc"
    """

    AI_INPUT_MESSAGES = "ai.input_messages"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_REQUEST_MESSAGES instead.

    The input messages to an LLM call.
    Example: [{"role": "user", "message": "hello"}]
    """

    AI_LOGIT_BIAS = "ai.logit_bias"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_* attributes instead.

    For an AI model call, the logit bias
    """

    AI_METADATA = "ai.metadata"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_* attributes instead.

    Extra metadata passed to an AI pipeline step.
    Example: {"executed_function": "add_integers"}
    """

    AI_MODEL_ID = "ai.model_id"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_REQUEST_MODEL or GEN_AI_RESPONSE_MODEL instead.

    The unique descriptor of the model being executed.
    Example: gpt-4
    """

    AI_PIPELINE_NAME = "ai.pipeline.name"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_PIPELINE_NAME instead.

    Name of the AI pipeline or chain being executed.
    Example: "qa-pipeline"
    """

    AI_PREAMBLE = "ai.preamble"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_* attributes instead.

    For an AI model call, the preamble parameter.
    Preambles are a part of the prompt used to adjust the model's overall behavior and conversation style.
    Example: "You are now a clown."
    """

    AI_PRESENCE_PENALTY = "ai.presence_penalty"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_REQUEST_PRESENCE_PENALTY instead.

    Used to reduce repetitiveness of generated tokens.
    Example: 0.5
    """

    AI_RAW_PROMPTING = "ai.raw_prompting"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_* attributes instead.

    Minimize pre-processing done to the prompt sent to the LLM.
    Example: true
    """

    AI_RESPONSE_FORMAT = "ai.response_format"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_* attributes instead.

    For an AI model call, the format of the response
    """

    AI_RESPONSES = "ai.responses"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_RESPONSE_TEXT instead.

    The responses to an AI model call. Always as a list.
    Example: ["hello", "world"]
    """

    AI_SEARCH_QUERIES = "ai.search_queries"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_* attributes instead.

    Queries used to search for relevant context or documents.
    Example: ["climate change effects", "renewable energy"]
    """

    AI_SEARCH_REQUIRED = "ai.is_search_required"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_* attributes instead.

    Boolean indicating if the model needs to perform a search.
    Example: true
    """

    AI_SEARCH_RESULTS = "ai.search_results"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_* attributes instead.

    Results returned from search queries for context.
    Example: ["Result 1", "Result 2"]
    """

    AI_SEED = "ai.seed"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_REQUEST_SEED instead.

    The seed, ideally models given the same seed and same other parameters will produce the exact same output.
    Example: 123.45
    """

    AI_STREAMING = "ai.streaming"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_RESPONSE_STREAMING instead.

    Whether or not the AI model call's response was streamed back asynchronously
    Example: true
    """

    AI_TAGS = "ai.tags"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_* attributes instead.

    Tags that describe an AI pipeline step.
    Example: {"executed_function": "add_integers"}
    """

    AI_TEMPERATURE = "ai.temperature"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_REQUEST_TEMPERATURE instead.

    For an AI model call, the temperature parameter. Temperature essentially means how random the output will be.
    Example: 0.5
    """

    AI_TEXTS = "ai.texts"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_* attributes instead.

    Raw text inputs provided to the model.
    Example: ["What is machine learning?"]
    """

    AI_TOP_K = "ai.top_k"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_REQUEST_TOP_K instead.

    For an AI model call, the top_k parameter. Top_k essentially controls how random the output will be.
    Example: 35
    """

    AI_TOP_P = "ai.top_p"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_REQUEST_TOP_P instead.

    For an AI model call, the top_p parameter. Top_p essentially controls how random the output will be.
    Example: 0.5
    """

    AI_TOOL_CALLS = "ai.tool_calls"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_RESPONSE_TOOL_CALLS instead.

    For an AI model call, the function that was called. This is deprecated for OpenAI, and replaced by tool_calls
    """

    AI_TOOLS = "ai.tools"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_REQUEST_AVAILABLE_TOOLS instead.

    For an AI model call, the functions that are available
    """

    AI_WARNINGS = "ai.warnings"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_* attributes instead.

    Warning messages generated during model execution.
    Example: ["Token limit exceeded"]
    """

    CACHE_HIT = "cache.hit"
    """
    A boolean indicating whether the requested data was found in the cache.
    Example: true
    """

    CACHE_ITEM_SIZE = "cache.item_size"
    """
    The size of the requested data in bytes.
    Example: 58
    """

    CACHE_KEY = "cache.key"
    """
    The key of the requested data.
    Example: template.cache.some_item.867da7e2af8e6b2f3aa7213a4080edb3
    """

    CLIENT_ADDRESS = "client.address"
    """
    Client address of the network connection - IP address or Unix domain socket name.
    Example: "10.1.2.80"
    """

    CODE_FILEPATH = "code.filepath"
    """
    .. deprecated::
        This attribute is deprecated. Use CODE_FILE_PATH instead.

    The source code file name that identifies the code unit as uniquely as possible (preferably an absolute file path).
    Example: "/app/myapplication/http/handler/server.py"
    """

    CODE_FILE_PATH = "code.file.path"
    """
    The source code file name that identifies the code unit as uniquely as possible (preferably an absolute file path).
    Example: "/app/myapplication/http/handler/server.py"
    """

    CODE_FUNCTION = "code.function"
    """
    .. deprecated::
        This attribute is deprecated. Use CODE_FUNCTION_NAME instead.

    The method or function name, or equivalent (usually rightmost part of the code unit's name).
    Example: "server_request"
    """

    CODE_FUNCTION_NAME = "code.function.name"
    """
    The method or function name, or equivalent (usually rightmost part of the code unit's name).
    Example: "server_request"
    """

    CODE_LINENO = "code.lineno"
    """
    .. deprecated::
        This attribute is deprecated. Use CODE_LINE_NUMBER instead.

    The line number in `code.filepath` best representing the operation. It SHOULD point within the code unit named in `code.function`.
    Example: 42
    """

    CODE_LINE_NUMBER = "code.line.number"
    """
    The line number in `code.file.path` best representing the operation. It SHOULD point within the code unit named in `code.function.name`.
    Example: 42
    """

    CODE_NAMESPACE = "code.namespace"
    """
    .. deprecated::
        This attribute is deprecated. Use CODE_FUNCTION_NAME instead; the namespace should be included within the function name.

    The "namespace" within which `code.function` is defined. Usually the qualified class or module name, such that `code.namespace` + some separator + `code.function` form a unique identifier for the code unit.
    Example: "http.handler"
    """

    DB_MONGODB_COLLECTION = "db.mongodb.collection"
    """
    The MongoDB collection being accessed within the database.
    See: https://github.com/open-telemetry/semantic-conventions/blob/main/docs/database/mongodb.md#attributes
    Example: public.users; customers
    """

    DB_NAME = "db.name"
    """
    .. deprecated::
        This attribute is deprecated. Use DB_NAMESPACE instead.

    The name of the database being accessed. For commands that switch the database, this should be set to the target database (even if the command fails).
    Example: myDatabase
    """

    DB_NAMESPACE = "db.namespace"
    """
    The name of the database being accessed.
    Example: "customers"
    """

    DB_DRIVER_NAME = "db.driver.name"
    """
    The name of the database driver being used for the connection.
    Example: "psycopg2"
    """

    DB_OPERATION = "db.operation"
    """
    .. deprecated::
        This attribute is deprecated. Use DB_OPERATION_NAME instead.

    The name of the operation being executed, e.g. the MongoDB command name such as findAndModify, or the SQL keyword.
    See: https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/trace/semantic_conventions/database.md
    Example: findAndModify, HMSET, SELECT
    """

    DB_OPERATION_NAME = "db.operation.name"
    """
    The name of the operation being executed.
    Example: "SELECT"
    """

    DB_SYSTEM = "db.system"
    """
    .. deprecated::
        This attribute is deprecated. Use DB_SYSTEM_NAME instead.

    An identifier for the database management system (DBMS) product being used.
    See: https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/trace/semantic_conventions/database.md
    Example: postgresql
    """

    DB_QUERY_TEXT = "db.query.text"
    """
    The database query being executed.
    Example: "SELECT * FROM users WHERE id = $1"
    """

    DB_SYSTEM_NAME = "db.system.name"
    """
    An identifier for the database management system (DBMS) product being used. See OpenTelemetry's list of well-known DBMS identifiers.
    Example: "postgresql"
    """

    DB_USER = "db.user"
    """
    The name of the database user used for connecting to the database.
    See: https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/trace/semantic_conventions/database.md
    Example: my_user
    """

    GEN_AI_AGENT_NAME = "gen_ai.agent.name"
    """
    The name of the agent being used.
    Example: "ResearchAssistant"
    """

    GEN_AI_CONVERSATION_ID = "gen_ai.conversation.id"
    """
    The unique identifier for the conversation/thread with the AI model.
    Example: "conv_abc123"
    """

    GEN_AI_CHOICE = "gen_ai.choice"
    """
    The model's response message.
    Example: "The weather in Paris is rainy and overcast, with temperatures around 57°F"
    """

    GEN_AI_EMBEDDINGS_INPUT = "gen_ai.embeddings.input"
    """
    The input to the embeddings operation.
    Example: "Hello!"
    """

    GEN_AI_FUNCTION_ID = "gen_ai.function_id"
    """
    Framework-specific tracing label for the execution of a function or other unit of execution in a generative AI system.
    Example: "my-awesome-function"
    """

    GEN_AI_OPERATION_NAME = "gen_ai.operation.name"
    """
    The name of the operation being performed.
    Example: "chat"
    """

    GEN_AI_PIPELINE_NAME = "gen_ai.pipeline.name"
    """
    Name of the AI pipeline or chain being executed.
    Example: "qa-pipeline"
    """

    GEN_AI_RESPONSE_FINISH_REASONS = "gen_ai.response.finish_reasons"
    """
    The reason why the model stopped generating.
    Example: "COMPLETE"
    """

    GEN_AI_RESPONSE_ID = "gen_ai.response.id"
    """
    Unique identifier for the completion.
    Example: "gen_123abc"
    """

    GEN_AI_RESPONSE_MODEL = "gen_ai.response.model"
    """
    Exact model identifier used to generate the response
    Example: gpt-4o-mini-2024-07-18
    """

    GEN_AI_RESPONSE_STREAMING = "gen_ai.response.streaming"
    """
    Whether or not the AI model call's response was streamed back asynchronously
    Example: true
    """

    GEN_AI_RESPONSE_TEXT = "gen_ai.response.text"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_OUTPUT_MESSAGES instead.

    The model's response text messages.
    Example: ["The weather in Paris is rainy and overcast, with temperatures around 57°F", "The weather in London is sunny and warm, with temperatures around 65°F"]
    """

    GEN_AI_OUTPUT_MESSAGES = "gen_ai.output.messages"
    """
    The model's response messages. It has to be a stringified version of an array of message objects, which can include text responses and tool calls.
    Example: [{"role": "assistant", "parts": [{"type": "text", "content": "The weather in Paris is currently rainy with a temperature of 57°F."}], "finish_reason": "stop"}]
    """

    GEN_AI_RESPONSE_TIME_TO_FIRST_TOKEN = "gen_ai.response.time_to_first_token"
    """
    The time it took to receive the first token from the model.
    Example: 0.1
    """

    GEN_AI_RESPONSE_TOOL_CALLS = "gen_ai.response.tool_calls"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_OUTPUT_MESSAGES instead.

    The tool calls in the model's response.
    Example: [{"name": "get_weather", "arguments": {"location": "Paris"}}]
    """

    GEN_AI_REQUEST_AVAILABLE_TOOLS = "gen_ai.request.available_tools"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_TOOL_DEFINITIONS instead.

    The available tools for the model.
    Example: [{"name": "get_weather", "description": "Get the weather for a given location"}, {"name": "get_news", "description": "Get the news for a given topic"}]
    """

    GEN_AI_TOOL_DEFINITIONS = "gen_ai.tool.definitions"
    """
    The list of source system tool definitions available to the GenAI agent or model.
    Example: [{"type": "function", "name": "get_current_weather", "description": "Get the current weather in a given location", "parameters": {"type": "object", "properties": {"location": {"type": "string", "description": "The city and state, e.g. San Francisco, CA"}, "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]}}, "required": ["location", "unit"]}}]
    """

    GEN_AI_REQUEST_FREQUENCY_PENALTY = "gen_ai.request.frequency_penalty"
    """
    The frequency penalty parameter used to reduce repetitiveness of generated tokens.
    Example: 0.1
    """

    GEN_AI_REQUEST_MAX_TOKENS = "gen_ai.request.max_tokens"
    """
    The maximum number of tokens to generate in the response.
    Example: 2048
    """

    GEN_AI_SYSTEM_INSTRUCTIONS = "gen_ai.system_instructions"
    """
    The system instructions passed to the model.
    Example: [{"type": "text", "text": "You are a helpful assistant."},{"type": "text", "text": "Be concise and clear."}]
    """

    GEN_AI_REQUEST_MESSAGES = "gen_ai.request.messages"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_INPUT_MESSAGES instead.

    The messages passed to the model. The "content" can be a string or an array of objects.
    Example: [{role: "system", "content: "Generate a random number."}, {"role": "user", "content": [{"text": "Generate a random number between 0 and 10.", "type": "text"}]}]
    """

    GEN_AI_INPUT_MESSAGES = "gen_ai.input.messages"
    """
    The messages passed to the model. It has to be a stringified version of an array of objects. Role values must be "user", "assistant", "tool", or "system".
    Example: [{"role": "user", "parts": [{"type": "text", "content": "Weather in Paris?"}]}, {"role": "assistant", "parts": [{"type": "tool_call", "id": "call_VSPygqKTWdrhaFErNvMV18Yl", "name": "get_weather", "arguments": {"location": "Paris"}}]}, {"role": "tool", "parts": [{"type": "tool_call_response", "id": "call_VSPygqKTWdrhaFErNvMV18Yl", "result": "rainy, 57°F"}]}]
    """

    GEN_AI_REQUEST_MODEL = "gen_ai.request.model"
    """
    The model identifier being used for the request.
    Example: "gpt-4-turbo"
    """

    GEN_AI_REQUEST_PRESENCE_PENALTY = "gen_ai.request.presence_penalty"
    """
    The presence penalty parameter used to reduce repetitiveness of generated tokens.
    Example: 0.1
    """

    GEN_AI_REQUEST_SEED = "gen_ai.request.seed"
    """
    The seed, ideally models given the same seed and same other parameters will produce the exact same output.
    Example: "1234567890"
    """

    GEN_AI_REQUEST_TEMPERATURE = "gen_ai.request.temperature"
    """
    The temperature parameter used to control randomness in the output.
    Example: 0.7
    """

    GEN_AI_REQUEST_TOP_K = "gen_ai.request.top_k"
    """
    Limits the model to only consider the K most likely next tokens, where K is an integer (e.g., top_k=20 means only the 20 highest probability tokens are considered).
    Example: 35
    """

    GEN_AI_REQUEST_TOP_P = "gen_ai.request.top_p"
    """
    The top_p parameter used to control diversity via nucleus sampling.
    Example: 1.0
    """

    GEN_AI_SYSTEM = "gen_ai.system"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_PROVIDER_NAME instead.

    The name of the AI system being used.
    Example: "openai"
    """

    GEN_AI_PROVIDER_NAME = "gen_ai.provider.name"
    """
    The Generative AI provider as identified by the client or server instrumentation.
    Example: "openai"
    """

    GEN_AI_TOOL_DESCRIPTION = "gen_ai.tool.description"
    """
    The description of the tool being used.
    Example: "Searches the web for current information about a topic"
    """

    GEN_AI_TOOL_INPUT = "gen_ai.tool.input"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_TOOL_CALL_ARGUMENTS instead.

    The input of the tool being used.
    Example: {"location": "Paris"}
    """

    GEN_AI_TOOL_CALL_ARGUMENTS = "gen_ai.tool.call.arguments"
    """
    The arguments of the tool call. It has to be a stringified version of the arguments to the tool.
    Example: {"location": "Paris"}
    """

    GEN_AI_TOOL_NAME = "gen_ai.tool.name"
    """
    The name of the tool being used.
    Example: "web_search"
    """

    GEN_AI_TOOL_OUTPUT = "gen_ai.tool.output"
    """
    .. deprecated::
        This attribute is deprecated. Use GEN_AI_TOOL_CALL_RESULT instead.

    The output of the tool being used.
    Example: "rainy, 57°F"
    """

    GEN_AI_TOOL_CALL_RESULT = "gen_ai.tool.call.result"
    """
    The result of the tool call. It has to be a stringified version of the result of the tool.
    Example: "rainy, 57°F"
    """

    GEN_AI_USAGE_INPUT_TOKENS = "gen_ai.usage.input_tokens"
    """
    The number of tokens in the input.
    Example: 150
    """

    GEN_AI_USAGE_INPUT_TOKENS_CACHED = "gen_ai.usage.input_tokens.cached"
    """
    The number of cached tokens in the input.
    Example: 50
    """

    GEN_AI_USAGE_INPUT_TOKENS_CACHE_WRITE = "gen_ai.usage.input_tokens.cache_write"
    """
    The number of tokens written to the cache when processing the AI input (prompt).
    Example: 100
    """

    GEN_AI_USAGE_OUTPUT_TOKENS = "gen_ai.usage.output_tokens"
    """
    The number of tokens in the output.
    Example: 250
    """

    GEN_AI_USAGE_OUTPUT_TOKENS_REASONING = "gen_ai.usage.output_tokens.reasoning"
    """
    The number of tokens used for reasoning in the output.
    Example: 75
    """

    GEN_AI_USAGE_TOTAL_TOKENS = "gen_ai.usage.total_tokens"
    """
    The total number of tokens used (input + output).
    Example: 400
    """

    GEN_AI_USER_MESSAGE = "gen_ai.user.message"
    """
    The user message passed to the model.
    Example: "What's the weather in Paris?"
    """

    HTTP_FRAGMENT = "http.fragment"
    """
    The Fragments present in the URL.
    Example: #foo=bar
    """

    HTTP_METHOD = "http.method"
    """
    .. deprecated::
        This attribute is deprecated. Use HTTP_REQUEST_METHOD instead.

    The HTTP method used.
    Example: GET
    """

    HTTP_REQUEST_BODY_DATA = "http.request.body.data"
    """
    HTTP request body data. Can be given as string or structural data of any format.
    Example: "[{\"role\": \"user\", \"message\": \"hello\"}]"
    """

    HTTP_REQUEST_HEADER = "http.request.header"
    """
    Prefix for HTTP request header attributes. The header name (lowercased) is
    appended to form the full attribute key.
    Example: "http.request.header.content-type"
    """

    HTTP_REQUEST_METHOD = "http.request.method"
    """
    The HTTP method used.
    Example: GET
    """

    HTTP_QUERY = "http.query"
    """
    The Query string present in the URL.
    Example: ?foo=bar&bar=baz
    """

    HTTP_STATUS_CODE = "http.response.status_code"
    """
    The HTTP status code as an integer.
    Example: 418
    """

    MESSAGING_DESTINATION_NAME = "messaging.destination.name"
    """
    The destination name where the message is being consumed from,
    e.g. the queue name or topic.
    """

    MESSAGING_MESSAGE_ID = "messaging.message.id"
    """
    The message's identifier.
    """

    MESSAGING_MESSAGE_RECEIVE_LATENCY = "messaging.message.receive.latency"
    """
    The latency between when the task was enqueued and when it was started to be processed.
    """

    MESSAGING_MESSAGE_RETRY_COUNT = "messaging.message.retry.count"
    """
    Number of retries/attempts to process a message.
    """

    MESSAGING_SYSTEM = "messaging.system"
    """
    The messaging system's name, e.g. `kafka`, `aws_sqs`
    """

    MIDDLEWARE_NAME = "middleware.name"
    """
    The middleware's name, e.g. `AuthenticationMiddleware`
    """

    NETWORK_PROTOCOL_NAME = "network.protocol.name"
    """
    The application layer protocol name used for the network connection.
    Example: "http", "https"
    """

    NETWORK_PEER_ADDRESS = "network.peer.address"
    """
    Peer address of the network connection - IP address or Unix domain socket name.
    Example: 10.1.2.80, /tmp/my.sock, localhost
    """

    NETWORK_PEER_PORT = "network.peer.port"
    """
    Peer port number of the network connection.
    Example: 6379
    """

    NETWORK_TRANSPORT = "network.transport"
    """
    The transport protocol used for the network connection.
    Example: "tcp", "udp", "unix"
    """

    PROCESS_PID = "process.pid"
    """
    The process ID of the running process.
    Example: 12345
    """

    PROCESS_COMMAND_ARGS = "process.command_args"
    """
    All the command arguments (including the command/executable itself) as received by the process.
    Example: ["cmd/otecol","--config=config.yaml"]
    """

    PROFILER_ID = "profiler_id"
    """
    Label identifying the profiler id that the span occurred in. This should be a string.
    Example: "5249fbada8d5416482c2f6e47e337372"
    """

    RPC_METHOD = "rpc.method"
    """
    The fully-qualified logical name of the method from the RPC interface perspective.
    Example: "com.example.ExampleService/exampleMethod"
    """

    RPC_RESPONSE_STATUS_CODE = "rpc.response.status_code"
    """
    Status code of the RPC returned by the RPC server or generated by the client.
    Example: "DEADLINE_EXCEEDED"
    """

    SERVER_ADDRESS = "server.address"
    """
    Name of the database host.
    Example: example.com
    """

    SERVER_PORT = "server.port"
    """
    Logical server port number
    Example: 80; 8080; 443
    """

    SERVER_SOCKET_ADDRESS = "server.socket.address"
    """
    Physical server IP address or Unix socket address.
    Example: 10.5.3.2
    """

    SERVER_SOCKET_PORT = "server.socket.port"
    """
    Physical server port.
    Recommended: If different than server.port.
    Example: 16456
    """

    THREAD_ID = "thread.id"
    """
    Identifier of a thread from where the span originated. This should be a string.
    Example: "7972576320"
    """

    THREAD_NAME = "thread.name"
    """
    Label identifying a thread from where the span originated. This should be a string.
    Example: "MainThread"
    """

    USER_EMAIL = "user.email"
    """
    User email address.
    Example: "test@example.com"
    """

    USER_ID = "user.id"
    """
    Unique identifier of the user.
    Example: "S-1-5-21-202424912787-2692429404-2351956786-1000"
    """

    USER_IP_ADDRESS = "user.ip_address"
    """
    The IP address of the user that triggered the request.
    Example: "10.1.2.80"
    """

    USER_NAME = "user.name"
    """
    Short name or login/username of the user.
    Example: "j.smith"
    """

    URL_FULL = "url.full"
    """
    The URL of the resource that was fetched.
    Example: "https://example.com/test?foo=bar#buzz"
    """

    URL_FRAGMENT = "url.fragment"
    """
    The fragments present in the URI. Note that this does not contain the leading # character, while the `http.fragment` attribute does.
    Example: "details"
    """

    URL_PATH = "url.path"
    """
    The URI path component.
    Example: "/foo"
    """

    URL_QUERY = "url.query"
    """
    The query string present in the URL. Note that this does not contain the leading ? character, while the `http.query` attribute does.
    Example: "foo=bar&bar=baz"
    """

    MCP_TOOL_NAME = "mcp.tool.name"
    """
    The name of the MCP tool being called.
    Example: "get_weather"
    """

    MCP_PROMPT_NAME = "mcp.prompt.name"
    """
    The name of the MCP prompt being retrieved.
    Example: "code_review"
    """

    MCP_RESOURCE_URI = "mcp.resource.uri"
    """
    The URI of the MCP resource being accessed.
    Example: "file:///path/to/resource"
    """

    MCP_METHOD_NAME = "mcp.method.name"
    """
    The MCP protocol method name being called.
    Example: "tools/call", "prompts/get", "resources/read"
    """

    MCP_REQUEST_ID = "mcp.request.id"
    """
    The unique identifier for the MCP request.
    Example: "req_123abc"
    """

    MCP_TOOL_RESULT_CONTENT = "mcp.tool.result.content"
    """
    The result/output content from an MCP tool execution.
    Example: "The weather is sunny"
    """

    MCP_TOOL_RESULT_CONTENT_COUNT = "mcp.tool.result.content_count"
    """
    The number of items/keys in the MCP tool result.
    Example: 5
    """

    MCP_TOOL_RESULT_IS_ERROR = "mcp.tool.result.is_error"
    """
    Whether the MCP tool execution resulted in an error.
    Example: True
    """

    MCP_PROMPT_RESULT_MESSAGE_CONTENT = "mcp.prompt.result.message_content"
    """
    The message content from an MCP prompt retrieval.
    Example: "Review the following code..."
    """

    MCP_PROMPT_RESULT_MESSAGE_ROLE = "mcp.prompt.result.message_role"
    """
    The role of the message in an MCP prompt retrieval (only set for single-message prompts).
    Example: "user", "assistant", "system"
    """

    MCP_PROMPT_RESULT_MESSAGE_COUNT = "mcp.prompt.result.message_count"
    """
    The number of messages in an MCP prompt result.
    Example: 1, 3
    """

    MCP_RESOURCE_PROTOCOL = "mcp.resource.protocol"
    """
    The protocol/scheme of the MCP resource URI.
    Example: "file", "http", "https"
    """

    MCP_TRANSPORT = "mcp.transport"
    """
    The transport method used for MCP communication.
    Example: "http", "sse", "stdio"
    """

    MCP_SESSION_ID = "mcp.session.id"
    """
    The session identifier for the MCP connection.
    Example: "a1b2c3d4e5f6"
    """

    SENTRY_DIST = "sentry.dist"
    """
    The Sentry dist.
    Example: "1.0"
    """

    SENTRY_ENVIRONMENT = "sentry.environment"
    """
    The Sentry environment.
    Example: "prod"
    """

    SENTRY_RELEASE = "sentry.release"
    """
    The Sentry release.
    Example: "1.2.3"
    """

    SENTRY_PLATFORM = "sentry.platform"
    """
    The sdk platform that generated the event.
    Example: "python"
    """

    SENTRY_SDK_NAME = "sentry.sdk.name"
    """
    The name of the SDK.
    Example: "python"
    """

    SENTRY_SDK_VERSION = "sentry.sdk.version"
    """
    The SDK version.
    Example: "1.2.3"
    """

    SENTRY_SDK_INTEGRATIONS = "sentry.sdk.integrations"
    """
    A list of names identifying enabled integrations.
    Example: ["AtexitIntegration", "StdlibIntegration"]
    """


class SPANSTATUS:
    """
    The status of a Sentry span.

    See: https://develop.sentry.dev/sdk/event-payloads/contexts/#trace-context
    """

    ABORTED = "aborted"
    ALREADY_EXISTS = "already_exists"
    CANCELLED = "cancelled"
    DATA_LOSS = "data_loss"
    DEADLINE_EXCEEDED = "deadline_exceeded"
    FAILED_PRECONDITION = "failed_precondition"
    INTERNAL_ERROR = "internal_error"
    INVALID_ARGUMENT = "invalid_argument"
    NOT_FOUND = "not_found"
    OK = "ok"
    OUT_OF_RANGE = "out_of_range"
    PERMISSION_DENIED = "permission_denied"
    RESOURCE_EXHAUSTED = "resource_exhausted"
    UNAUTHENTICATED = "unauthenticated"
    UNAVAILABLE = "unavailable"
    UNIMPLEMENTED = "unimplemented"
    UNKNOWN_ERROR = "unknown_error"


class OP:
    ANTHROPIC_MESSAGES_CREATE = "ai.messages.create.anthropic"
    CACHE_GET = "cache.get"
    CACHE_PUT = "cache.put"
    COHERE_CHAT_COMPLETIONS_CREATE = "ai.chat_completions.create.cohere"
    COHERE_EMBEDDINGS_CREATE = "ai.embeddings.create.cohere"
    DB = "db"
    DB_CURSOR_ITERATOR = "db.cursor.iter"
    DB_CURSOR_FETCH = "db.cursor.fetch"
    DB_REDIS = "db.redis"
    EVENT_DJANGO = "event.django"
    FUNCTION = "function"
    FUNCTION_AWS = "function.aws"
    FUNCTION_GCP = "function.gcp"
    GEN_AI_CHAT = "gen_ai.chat"
    GEN_AI_CREATE_AGENT = "gen_ai.create_agent"
    GEN_AI_EMBEDDINGS = "gen_ai.embeddings"
    GEN_AI_EXECUTE_TOOL = "gen_ai.execute_tool"
    GEN_AI_TEXT_COMPLETION = "gen_ai.text_completion"
    GEN_AI_HANDOFF = "gen_ai.handoff"
    GEN_AI_INVOKE_AGENT = "gen_ai.invoke_agent"
    GEN_AI_RESPONSES = "gen_ai.responses"
    GRAPHQL_EXECUTE = "graphql.execute"
    GRAPHQL_MUTATION = "graphql.mutation"
    GRAPHQL_PARSE = "graphql.parse"
    GRAPHQL_RESOLVE = "graphql.resolve"
    GRAPHQL_SUBSCRIPTION = "graphql.subscription"
    GRAPHQL_QUERY = "graphql.query"
    GRAPHQL_VALIDATE = "graphql.validate"
    GRPC_CLIENT = "grpc.client"
    GRPC_SERVER = "grpc.server"
    HTTP_CLIENT = "http.client"
    HTTP_CLIENT_STREAM = "http.client.stream"
    HTTP_SERVER = "http.server"
    MIDDLEWARE_DJANGO = "middleware.django"
    MIDDLEWARE_LITESTAR = "middleware.litestar"
    MIDDLEWARE_LITESTAR_RECEIVE = "middleware.litestar.receive"
    MIDDLEWARE_LITESTAR_SEND = "middleware.litestar.send"
    MIDDLEWARE_STARLETTE = "middleware.starlette"
    MIDDLEWARE_STARLETTE_RECEIVE = "middleware.starlette.receive"
    MIDDLEWARE_STARLETTE_SEND = "middleware.starlette.send"
    MIDDLEWARE_STARLITE = "middleware.starlite"
    MIDDLEWARE_STARLITE_RECEIVE = "middleware.starlite.receive"
    MIDDLEWARE_STARLITE_SEND = "middleware.starlite.send"
    HUGGINGFACE_HUB_CHAT_COMPLETIONS_CREATE = (
        "ai.chat_completions.create.huggingface_hub"
    )
    QUEUE_PROCESS = "queue.process"
    QUEUE_PUBLISH = "queue.publish"
    QUEUE_SUBMIT_ARQ = "queue.submit.arq"
    QUEUE_TASK_ARQ = "queue.task.arq"
    QUEUE_SUBMIT_CELERY = "queue.submit.celery"
    QUEUE_TASK_CELERY = "queue.task.celery"
    QUEUE_TASK_RQ = "queue.task.rq"
    QUEUE_SUBMIT_HUEY = "queue.submit.huey"
    QUEUE_TASK_HUEY = "queue.task.huey"
    QUEUE_SUBMIT_RAY = "queue.submit.ray"
    QUEUE_TASK_RAY = "queue.task.ray"
    QUEUE_TASK_DRAMATIQ = "queue.task.dramatiq"
    QUEUE_SUBMIT_DJANGO = "queue.submit.django"
    SUBPROCESS = "subprocess"
    SUBPROCESS_WAIT = "subprocess.wait"
    SUBPROCESS_COMMUNICATE = "subprocess.communicate"
    TEMPLATE_RENDER = "template.render"
    VIEW_RENDER = "view.render"
    VIEW_RESPONSE_RENDER = "view.response.render"
    WEBSOCKET_SERVER = "websocket.server"
    SOCKET_CONNECTION = "socket.connection"
    SOCKET_DNS = "socket.dns"
    MCP_SERVER = "mcp.server"


# This type exists to trick mypy and PyCharm into thinking `init` and `Client`
# take these arguments (even though they take opaque **kwargs)
class ClientConstructor:
    def __init__(
        self,
        dsn: "Optional[str]" = None,
        *,
        max_breadcrumbs: int = DEFAULT_MAX_BREADCRUMBS,
        release: "Optional[str]" = None,
        environment: "Optional[str]" = None,
        server_name: "Optional[str]" = None,
        shutdown_timeout: float = 2,
        integrations: "Sequence[sentry_sdk.integrations.Integration]" = [],  # noqa: B006
        ignore_spans: "Optional[IgnoreSpansConfig]" = None,
        in_app_include: "List[str]" = [],  # noqa: B006
        in_app_exclude: "List[str]" = [],  # noqa: B006
        default_integrations: bool = True,
        dist: "Optional[str]" = None,
        transport: "Optional[Union[sentry_sdk.transport.Transport, Type[sentry_sdk.transport.Transport], Callable[[Event], None]]]" = None,
        transport_queue_size: int = DEFAULT_QUEUE_SIZE,
        sample_rate: float = 1.0,
        send_default_pii: "Optional[bool]" = None,
        http_proxy: "Optional[str]" = None,
        https_proxy: "Optional[str]" = None,
        ignore_errors: "Sequence[Union[type, str]]" = [],  # noqa: B006
        max_request_body_size: str = "medium",
        socket_options: "Optional[List[Tuple[int, int, int | bytes]]]" = None,
        keep_alive: "Optional[bool]" = None,
        before_send: "Optional[EventProcessor]" = None,
        before_breadcrumb: "Optional[BreadcrumbProcessor]" = None,
        debug: "Optional[bool]" = None,
        attach_stacktrace: bool = False,
        ca_certs: "Optional[str]" = None,
        propagate_traces: bool = True,
        traces_sample_rate: "Optional[float]" = None,
        trace_lifecycle: "Optional[Literal['static', 'stream']]" = None,
        traces_sampler: "Optional[TracesSampler]" = None,
        profiles_sample_rate: "Optional[float]" = None,
        profiles_sampler: "Optional[TracesSampler]" = None,
        profiler_mode: "Optional[ProfilerMode]" = None,
        profile_lifecycle: 'Literal["manual", "trace"]' = "manual",
        profile_session_sample_rate: "Optional[float]" = None,
        auto_enabling_integrations: bool = True,
        disabled_integrations: "Optional[Sequence[sentry_sdk.integrations.Integration]]" = None,
        auto_session_tracking: bool = True,
        send_client_reports: bool = True,
        _experiments: "Experiments" = {},  # noqa: B006
        proxy_headers: "Optional[Dict[str, str]]" = None,
        instrumenter: "Optional[str]" = INSTRUMENTER.SENTRY,
        before_send_transaction: "Optional[TransactionProcessor]" = None,
        project_root: "Optional[str]" = None,
        enable_tracing: "Optional[bool]" = None,
        include_local_variables: "Optional[bool]" = True,
        include_source_context: "Optional[bool]" = True,
        trace_propagation_targets: "Optional[Sequence[str]]" = [  # noqa: B006
            MATCH_ALL
        ],
        functions_to_trace: "Sequence[Dict[str, str]]" = [],  # noqa: B006
        event_scrubber: "Optional[sentry_sdk.scrubber.EventScrubber]" = None,
        max_value_length: "Optional[int]" = DEFAULT_MAX_VALUE_LENGTH,
        enable_backpressure_handling: bool = True,
        error_sampler: "Optional[Callable[[Event, Hint], Union[float, bool]]]" = None,
        enable_db_query_source: bool = True,
        db_query_source_threshold_ms: int = 100,
        enable_http_request_source: bool = True,
        http_request_source_threshold_ms: int = 100,
        spotlight: "Optional[Union[bool, str]]" = None,
        cert_file: "Optional[str]" = None,
        key_file: "Optional[str]" = None,
        custom_repr: "Optional[Callable[..., Optional[str]]]" = None,
        add_full_stack: bool = DEFAULT_ADD_FULL_STACK,
        max_stack_frames: "Optional[int]" = DEFAULT_MAX_STACK_FRAMES,
        enable_logs: bool = False,
        before_send_log: "Optional[Callable[[Log, Hint], Optional[Log]]]" = None,
        trace_ignore_status_codes: "AbstractSet[int]" = frozenset(),
        enable_metrics: bool = True,
        before_send_metric: "Optional[Callable[[Metric, Hint], Optional[Metric]]]" = None,
        org_id: "Optional[str]" = None,
        strict_trace_continuation: bool = False,
        stream_gen_ai_spans: bool = True,
    ) -> None:
        """Initialize the Sentry SDK with the given parameters. All parameters described here can be used in a call to `sentry_sdk.init()`.

        :param dsn: The DSN tells the SDK where to send the events.

            If this option is not set, the SDK will just not send any data.

            The `dsn` config option takes precedence over the environment variable.

            Learn more about `DSN utilization <https://docs.sentry.io/product/sentry-basics/dsn-explainer/#dsn-utilization>`_.

        :param debug: Turns debug mode on or off.

            When `True`, the SDK will attempt to print out debugging information. This can be useful if something goes
            wrong with event sending.

            The default is always `False`. It's generally not recommended to turn it on in production because of the
            increase in log output.

            The `debug` config option takes precedence over the environment variable.

        :param release: Sets the release.

            If not set, the SDK will try to automatically configure a release out of the box but it's a better idea to
            manually set it to guarantee that the release is in sync with your deploy integrations.

            Release names are strings, but some formats are detected by Sentry and might be rendered differently.

            See `the releases documentation <https://docs.sentry.io/platforms/python/configuration/releases/>`_ to learn how the SDK tries to
            automatically configure a release.

            The `release` config option takes precedence over the environment variable.

            Learn more about how to send release data so Sentry can tell you about regressions between releases and
            identify the potential source in `the product documentation <https://docs.sentry.io/product/releases/>`_.

        :param environment: Sets the environment. This string is freeform and set to `production` by default.

            A release can be associated with more than one environment to separate them in the UI (think `staging` vs
            `production` or similar).

            The `environment` config option takes precedence over the environment variable.

        :param dist: The distribution of the application.

            Distributions are used to disambiguate build or deployment variants of the same release of an application.

            The dist can be for example a build number.

        :param sample_rate: Configures the sample rate for error events, in the range of `0.0` to `1.0`.

            The default is `1.0`, which means that 100% of error events will be sent. If set to `0.1`, only 10% of
            error events will be sent.

            Events are picked randomly.

        :param error_sampler: Dynamically configures the sample rate for error events on a per-event basis.

            This configuration option accepts a function, which takes two parameters (the `event` and the `hint`), and
            which returns a boolean (indicating whether the event should be sent to Sentry) or a floating-point number
            between `0.0` and `1.0`, inclusive.

            The number indicates the probability the event is sent to Sentry; the SDK will randomly decide whether to
            send the event with the given probability.

            If this configuration option is specified, the `sample_rate` option is ignored.

        :param ignore_errors: A list of exception class names that shouldn't be sent to Sentry.

            Errors that are an instance of these exceptions or a subclass of them, will be filtered out before they're
            sent to Sentry.

            By default, all errors are sent.

        :param max_breadcrumbs: This variable controls the total amount of breadcrumbs that should be captured.

            This defaults to `100`, but you can set this to any number.

            However, you should be aware that Sentry has a `maximum payload size <https://develop.sentry.dev/sdk/data-model/envelopes/#size-limits>`_
            and any events exceeding that payload size will be dropped.

        :param attach_stacktrace: When enabled, stack traces are automatically attached to all messages logged.

            Stack traces are always attached to exceptions; however, when this option is set, stack traces are also
            sent with messages.

            This option means that stack traces appear next to all log messages.

            Grouping in Sentry is different for events with stack traces and without. As a result, you will get new
            groups as you enable or disable this flag for certain events.

        :param send_default_pii: If this flag is enabled, `certain personally identifiable information (PII)
            <https://docs.sentry.io/platforms/python/data-management/data-collected/>`_ is added by active integrations.

            If you enable this option, be sure to manually remove what you don't want to send using our features for
            managing `Sensitive Data <https://docs.sentry.io/data-management/sensitive-data/>`_.

        :param event_scrubber: Scrubs the event payload for sensitive information such as cookies, sessions, and
            passwords from a `denylist`.

            It can additionally be used to scrub from another `pii_denylist` if `send_default_pii` is disabled.

            See how to `configure the scrubber here <https://docs.sentry.io/data-management/sensitive-data/#event-scrubber>`_.

        :param include_source_context: When enabled, source context will be included in events sent to Sentry.

            This source context includes the five lines of code above and below the line of code where an error
            happened.

        :param include_local_variables: When enabled, the SDK will capture a snapshot of local variables to send with
            the event to help with debugging.

        :param add_full_stack: When capturing errors, Sentry stack traces typically only include frames that start the
            moment an error occurs.

            But if the `add_full_stack` option is enabled (set to `True`), all frames from the start of execution will
            be included in the stack trace sent to Sentry.

        :param max_stack_frames: This option limits the number of stack frames that will be captured when
            `add_full_stack` is enabled.

        :param server_name: This option can be used to supply a server name.

            When provided, the name of the server is sent along and persisted in the event.

            For many integrations, the server name actually corresponds to the device hostname, even in situations
            where the machine is not actually a server.

        :param project_root: The full path to the root directory of your application.

            The `project_root` is used to mark frames in a stack trace either as being in your application or outside
            of the application.

        :param in_app_include: A list of string prefixes of module names that belong to the app.

            This option takes precedence over `in_app_exclude`.

            Sentry differentiates stack frames that are directly related to your application ("in application") from
            stack frames that come from other packages such as the standard library, frameworks, or other dependencies.

            The application package is automatically marked as `inApp`.

            The difference is visible in [sentry.io](https://sentry.io), where only the "in application" frames are
            displayed by default.

        :param in_app_exclude: A list of string prefixes of module names that do not belong to the app, but rather to
            third-party packages.

            Modules considered not part of the app will be hidden from stack traces by default.

            This option can be overridden using `in_app_include`.

        :param max_request_body_size: This parameter controls whether integrations should capture HTTP request bodies.
            It can be set to one of the following values:

            - `never`: Request bodies are never sent.
            - `small`: Only small request bodies will be captured. The cutoff for small depends on the SDK (typically
              4KB).
            - `medium`: Medium and small requests will be captured (typically 10KB).
            - `always`: The SDK will always capture the request body as long as Sentry can make sense of it.

            Please note that the Sentry server [limits HTTP request body size](https://develop.sentry.dev/sdk/
            expected-features/data-handling/#variable-size). The server always enforces its size limit, regardless of
            how you configure this option.

        :param max_value_length: The number of characters after which the values containing text in the event payload
            will be truncated.

            WARNING: If the value you set for this is exceptionally large, the event may exceed 1 MiB and will be
            dropped by Sentry.

        :param ca_certs: A path to an alternative CA bundle file in PEM-format.

        :param send_client_reports: Set this boolean to `False` to disable sending of client reports.

            Client reports allow the client to send status reports about itself to Sentry, such as information about
            events that were dropped before being sent.

        :param integrations: List of integrations to enable in addition to `auto-enabling integrations (overview)
            <https://docs.sentry.io/platforms/python/integrations>`_.

            This setting can be used to override the default config options for a specific auto-enabling integration
            or to add an integration that is not auto-enabled.

        :param disabled_integrations: List of integrations that will be disabled.

            This setting can be used to explicitly turn off specific `auto-enabling integrations (list)
            <https://docs.sentry.io/platforms/python/integrations/#available-integrations>`_ or
            `default <https://docs.sentry.io/platforms/python/integrations/default-integrations/>`_ integrations.

        :param auto_enabling_integrations: Configures whether `auto-enabling integrations (configuration)
            <https://docs.sentry.io/platforms/python/integrations/#available-integrations>`_ should be enabled.

            When set to `False`, no auto-enabling integrations will be enabled by default, even if the corresponding
            framework/library is detected.

        :param default_integrations: Configures whether `default integrations
            <https://docs.sentry.io/platforms/python/integrations/default-integrations/>`_ should be enabled.

            Setting `default_integrations` to `False` disables all default integrations **as well as all auto-enabling
            integrations**, unless they are specifically added in the `integrations` option, described above.

        :param before_send: This function is called with an SDK-specific message or error event object, and can return
            a modified event object, or `null` to skip reporting the event.

            This can be used, for instance, for manual PII stripping before sending.

            By the time `before_send` is executed, all scope data has already been applied to the event. Further
            modification of the scope won't have any effect.

        :param before_send_transaction: This function is called with an SDK-specific transaction event object, and can
            return a modified transaction event object, or `null` to skip reporting the event.

            One way this might be used is for manual PII stripping before sending.

        :param before_breadcrumb: This function is called with an SDK-specific breadcrumb object before the breadcrumb
            is added to the scope.

            When nothing is returned from the function, the breadcrumb is dropped.

            To pass the breadcrumb through, return the first argument, which contains the breadcrumb object.

            The callback typically gets a second argument (called a "hint") which contains the original object from
            which the breadcrumb was created to further customize what the breadcrumb should look like.

        :param transport: Switches out the transport used to send events.

            How this works depends on the SDK. It can, for instance, be used to capture events for unit-testing or to
            send it through some more complex setup that requires proxy authentication.

        :param transport_queue_size: The maximum number of events that will be queued before the transport is forced to
            flush.

        :param http_proxy: When set, a proxy can be configured that should be used for outbound requests.

            This is also used for HTTPS requests unless a separate `https_proxy` is configured. However, not all SDKs
            support a separate HTTPS proxy.

            SDKs will attempt to default to the system-wide configured proxy, if possible. For instance, on Unix
            systems, the `http_proxy` environment variable will be picked up.

        :param https_proxy: Configures a separate proxy for outgoing HTTPS requests.

            This value might not be supported by all SDKs. When not supported the `http-proxy` value is also used for
            HTTPS requests at all times.

        :param proxy_headers: A dict containing additional proxy headers (usually for authentication) to be forwarded
            to `urllib3`'s `ProxyManager <https://urllib3.readthedocs.io/en/1.24.3/reference/index.html#urllib3.poolmanager.ProxyManager>`_.

        :param shutdown_timeout: Controls how many seconds to wait before shutting down.

            Sentry SDKs send events from a background queue. This queue is given a certain amount to drain pending
            events. The default is SDK specific but typically around two seconds.

            Setting this value too low may cause problems for sending events from command line applications.

            Setting the value too high will cause the application to block for a long time for users experiencing
            network connectivity problems.

        :param keep_alive: Determines whether to keep the connection alive between requests.

            This can be useful in environments where you encounter frequent network issues such as connection resets.

        :param cert_file: Path to the client certificate to use.

            If set, supersedes the `CLIENT_CERT_FILE` environment variable.

        :param key_file: Path to the key file to use.

            If set, supersedes the `CLIENT_KEY_FILE` environment variable.

        :param socket_options: An optional list of socket options to use.

            These provide fine-grained, low-level control over the way the SDK connects to Sentry.

            If provided, the options will override the default `urllib3` `socket options
            <https://urllib3.readthedocs.io/en/stable/reference/urllib3.connection.html#urllib3.connection.HTTPConnection>`_.

        :param traces_sample_rate: A number between `0` and `1`, controlling the percentage chance a given transaction
            will be sent to Sentry.

            (`0` represents 0% while `1` represents 100%.) Applies equally to all transactions created in the app.

            Either this or `traces_sampler` must be defined to enable tracing.

            If `traces_sample_rate` is `0`, this means that no new traces will be created. However, if you have
            another service (for example a JS frontend) that makes requests to your service that include trace
            information, those traces will be continued and thus transactions will be sent to Sentry.

            If you want to disable all tracing you need to set `traces_sample_rate=None`. In this case, no new traces
            will be started and no incoming traces will be continued.

        :param traces_sampler: A function responsible for determining the percentage chance a given transaction will be
            sent to Sentry.

            It will automatically be passed information about the transaction and the context in which it's being
            created, and must return a number between `0` (0% chance of being sent) and `1` (100% chance of being
            sent).

            Can also be used for filtering transactions, by returning `0` for those that are unwanted.

            Either this or `traces_sample_rate` must be defined to enable tracing.

        :param trace_propagation_targets: An optional property that controls which downstream services receive tracing
            data, in the form of a `sentry-trace` and a `baggage` header attached to any outgoing HTTP requests.

            The option may contain a list of strings or regex against which the URLs of outgoing requests are matched.

            If one of the entries in the list matches the URL of an outgoing request, trace data will be attached to
            that request.

            String entries do not have to be full matches, meaning the URL of a request is matched when it _contains_
            a string provided through the option.

            If `trace_propagation_targets` is not provided, trace data is attached to every outgoing request from the
            instrumented client.

        :param functions_to_trace: An optional list of functions that should be set up for tracing.

            For each function in the list, a span will be created when the function is executed.

            Functions in the list are represented as strings containing the fully qualified name of the function.

            This is a convenient option, making it possible to have one central place for configuring what functions
            to trace, instead of having custom instrumentation scattered all over your code base.

            To learn more, see the `Custom Instrumentation <https://docs.sentry.io/platforms/python/tracing/instrumentation/custom-instrumentation/#define-span-creation-in-a-central-place>`_ documentation.

        :param enable_backpressure_handling: When enabled, a new monitor thread will be spawned to perform health
            checks on the SDK.

            If the system is unhealthy, the SDK will keep halving the `traces_sample_rate` set by you in 10 second
            intervals until recovery.

            This down sampling helps ensure that the system stays stable and reduces SDK overhead under high load.

            This option is enabled by default.

        :param enable_db_query_source: When enabled, the source location will be added to database queries.

        :param db_query_source_threshold_ms: The threshold in milliseconds for adding the source location to database
            queries.

            The query location will be added to the query for queries slower than the specified threshold.

        :param enable_http_request_source: When enabled, the source location will be added to outgoing HTTP requests.

        :param http_request_source_threshold_ms: The threshold in milliseconds for adding the source location to an
            outgoing HTTP request.

            The request location will be added to the request for requests slower than the specified threshold.

        :param custom_repr: A custom `repr <https://docs.python.org/3/library/functions.html#repr>`_ function to run
            while serializing an object.

            Use this to control how your custom objects and classes are visible in Sentry.

            Return a string for that repr value to be used or `None` to continue serializing how Sentry would have
            done it anyway.

        :param profiles_sample_rate: A number between `0` and `1`, controlling the percentage chance a given sampled
            transaction will be profiled.

            (`0` represents 0% while `1` represents 100%.) Applies equally to all transactions created in the app.

            This is relative to the tracing sample rate - e.g. `0.5` means 50% of sampled transactions will be
            profiled.

        :param profiles_sampler:

        :param profiler_mode:

        :param profile_lifecycle:

        :param profile_session_sample_rate:

        :param enable_tracing:

        :param propagate_traces:

        :param auto_session_tracking:

        :param spotlight:

        :param instrumenter:

        :param enable_logs: Set `enable_logs` to True to enable the SDK to emit
            Sentry logs. Defaults to False.

        :param before_send_log: An optional function to modify or filter out logs
            before they're sent to Sentry. Any modifications to the log in this
            function will be retained. If the function returns None, the log will
            not be sent to Sentry.

        :param trace_ignore_status_codes: An optional property that disables tracing for
            HTTP requests with certain status codes.

            Requests are not traced if the status code is contained in the provided set.

            If `trace_ignore_status_codes` is not provided, requests with any status code
            may be traced.

            This option has no effect in span streaming mode (`trace_lifecycle="stream"`).

        :param strict_trace_continuation: If set to `True`, the SDK will only continue a trace if the `org_id` of the incoming trace found in the
           `baggage` header matches the `org_id` of the current Sentry client and only if BOTH are present.

            If set to `False`, consistency of `org_id` will only be enforced if both are present. If either are missing, the trace will be continued.

            The client's organization ID is extracted from the DSN or can be set with the `org_id` option.
            If the organization IDs do not match, the SDK will start a new trace instead of continuing the incoming one.
            This is useful to prevent traces of unknown third-party services from being continued in your application.

        :param org_id: An optional organization ID. The SDK will try to extract if from the DSN in most cases
            but you can provide it explicitly for self-hosted and Relay setups. This value is used for
            trace propagation and for features like `strict_trace_continuation`.

        :param stream_gen_ai_spans: When set, generative AI spans are sent in a new transport format to
            reduce downstream data loss.

        :param trace_lifecycle: Controls how traces are sent. Set to `"stream"` to send spans as they
            finish, or `"static"` to send a completed trace as a transaction event.

        :param ignore_spans: A sequence of span-matching rules. Matching spans are ignored when
            `trace_lifecycle="stream"` is enabled.

        :param _experiments: Dictionary of experimental, opt-in features that are not yet stable.

            ``data_collection`` (EXPERIMENTAL): structured configuration controlling what data integrations
            collect automatically, superseding `send_default_pii`. Passing a dict under
            `_experiments={"data_collection": {...}}` opts into the feature; omitted fields use their
            defaults (most categories are collected, with the sensitive denylist scrubbing values).
            When it is not set, the SDK derives behaviour from `send_default_pii` so that upgrading
            changes nothing. Restrict collection per category (user identity, cookies, HTTP
            headers/bodies, query params, generative AI inputs/outputs, stack frame variables, source
            context). If `send_default_pii` is also set, `data_collection` takes precedence.

            Example::

                sentry_sdk.init(
                    dsn="...",
                    _experiments={"data_collection": {"user_info": False, "http_bodies": []}},
                )

            See https://docs.sentry.io/platforms/python/configuration/options/#data_collection for more details.
        """
        pass


def _get_default_options() -> "dict[str, Any]":
    import inspect

    a = inspect.getfullargspec(ClientConstructor.__init__)
    defaults = a.defaults or ()
    kwonlydefaults = a.kwonlydefaults or {}

    return dict(
        itertools.chain(
            zip(a.args[-len(defaults) :], defaults),
            kwonlydefaults.items(),
        )
    )


DEFAULT_OPTIONS = _get_default_options()
del _get_default_options


VERSION = "2.66.1"
